@@ -17,18 +17,13 @@
  *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include <grub/normal.h>
 #include <grub/dl.h>
-#include <grub/machine/machine.h>
-
-#if defined(GRUB_MACHINE_IEEE1275)
-#include <grub/machine/kernel.h>
-#elif defined(GRUB_MACHINE_EFI)
-#include <grub/efi/efi.h>
-#endif
+#include <grub/command.h>
+#include <grub/misc.h>
+#include <grub/i18n.h>
 
 static grub_err_t
-grub_cmd_halt (struct grub_arg_list *state __attribute__ ((unused)),
+grub_cmd_halt (grub_command_t cmd __attribute__ ((unused)),
 	       int argc __attribute__ ((unused)),
 	       char **args __attribute__ ((unused)))
 {
@@ -36,16 +31,16 @@ grub_cmd_halt (struct grub_arg_list *state __attribute__ ((unused)),
   return 0;
 }
 
+static grub_command_t cmd;
 
 GRUB_MOD_INIT(halt)
 {
-  (void)mod;			/* To stop warning. */
-  grub_register_command ("halt", grub_cmd_halt, GRUB_COMMAND_FLAG_BOTH,
-			 "halt", "halts the computer.  This command does not"
-			 " work on all firmware.", 0);
+  cmd = grub_register_command ("halt", grub_cmd_halt,
+			       0, N_("Halts the computer.  This command does" 
+			       " not work on all firmware implementations."));
 }
 
 GRUB_MOD_FINI(halt)
 {
-  grub_unregister_command ("halt");
+  grub_unregister_command (cmd);
 }
