@@ -1,7 +1,7 @@
 /* halt.c - command to halt the computer.  */
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 2005,2007  Free Software Foundation, Inc.
+ *  Copyright (C) 2005,2007,2009  Free Software Foundation, Inc.
  *
  *  GRUB is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -17,23 +17,24 @@
  *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include <grub/normal.h>
 #include <grub/dl.h>
-#include <grub/arg.h>
-#include <grub/machine/init.h>
+#include <grub/misc.h>
+#include <grub/extcmd.h>
+#include <grub/i18n.h>
 
 static const struct grub_arg_option options[] =
   {
-    {"no-apm", 'n', 0, "do not use APM to halt the computer", 0, 0},
+    {"no-apm", 'n', 0, N_("Do not use APM to halt the computer."), 0, 0},
     {0, 0, 0, 0, 0, 0}
   };
 
 static grub_err_t
-grub_cmd_halt (struct grub_arg_list *state,
+grub_cmd_halt (grub_extcmd_t cmd,
 	       int argc __attribute__ ((unused)),
 	       char **args __attribute__ ((unused)))
 
 {
+  struct grub_arg_list *state = cmd->state;
   int no_apm = 0;
   if (state[0].set)
     no_apm = 1;
@@ -41,17 +42,17 @@ grub_cmd_halt (struct grub_arg_list *state,
   return 0;
 }
 
-
+static grub_extcmd_t cmd;
 
 GRUB_MOD_INIT(halt)
 {
-  (void)mod;			/* To stop warning. */
-  grub_register_command ("halt", grub_cmd_halt, GRUB_COMMAND_FLAG_BOTH,
-			 "halt [-n]",
-			 "Halt the system, if possible using APM", options);
+  cmd = grub_register_extcmd ("halt", grub_cmd_halt, GRUB_COMMAND_FLAG_BOTH,
+			      "[-n]",
+			      N_("Halt the system, if possible using APM."),
+			      options);
 }
 
 GRUB_MOD_FINI(halt)
 {
-  grub_unregister_command ("halt");
+  grub_unregister_extcmd (cmd);
 }
