@@ -1,7 +1,7 @@
 /* ofdisk.c - Open Firmware disk access.  */
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 2004,2006,2007  Free Software Foundation, Inc.
+ *  Copyright (C) 2004,2006,2007,2008,2009  Free Software Foundation, Inc.
  *
  *  GRUB is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -23,6 +23,53 @@
 #include <grub/ieee1275/ieee1275.h>
 #include <grub/ieee1275/ofdisk.h>
 
+struct ofdisk_hash_ent
+{
+  char *devpath;
+  struct ofdisk_hash_ent *next;
+};
+
+#define OFDISK_HASH_SZ	8
+static struct ofdisk_hash_ent *ofdisk_hash[OFDISK_HASH_SZ];
+
+static int
+ofdisk_hash_fn (const char *devpath)
+{
+  int hash = 0;
+  while (*devpath)
+    hash ^= *devpath++;
+  return (hash & (OFDISK_HASH_SZ - 1));
+}
+
+static struct ofdisk_hash_ent *
+ofdisk_hash_find (const char *devpath)
+{
+  struct ofdisk_hash_ent *p = ofdisk_hash[ofdisk_hash_fn(devpath)];
+
+  while (p)
+    {
+      if (!grub_strcmp (p->devpath, devpath))
+	break;
+      p = p->next;
+    }
+  return p;
+}
+
+static struct ofdisk_hash_ent *
+ofdisk_hash_add (char *devpath)
+{
+  struct ofdisk_hash_ent **head = &ofdisk_hash[ofdisk_hash_fn(devpath)];
+  struct ofdisk_hash_ent *p = grub_malloc(sizeof (*p));
+
+  if (p)
+    {
+      p->devpath = devpath;
+      p->next = *head;
+      *head = p;
+    }
+  return p;
+}
+
 static int
 grub_ofdisk_iterate (int (*hook) (const char *name))
 {
@@ -30,6 +77,8 @@ grub_ofdisk_iterate (int (*hook) (const char *name))
 
   int dev_iterate (struct grub_ieee1275_devalias *alias)
     {
+      int ret = 0;
+
       grub_dprintf ("disk", "disk name = %s\n", alias->name);
 
       if (grub_ieee1275_test_flag (GRUB_IEEE1275_FLAG_OFDISK_SDCARD_ONLY))
@@ -57,18 +106,45 @@ grub_ofdisk_iterate (int (*hook) (const char *name))
 	    }
 	}
 
-      if (! grub_strcmp (alias->type, "block"))
-	hook (alias->name);
-      else if ((! grub_strcmp (alias->type, "scsi"))
-	       || (! grub_strcmp (alias->type, "ide"))
-	       || (! grub_strcmp (alias->type, "ata")))
-	/* Search for block-type children of these bus controllers.  */
-	grub_children_iterate (alias->name, dev_iterate);
-      return 0;
+      if (! grub_strcmp (alias->type, "block") &&
+	  grub_strncmp (alias->name, "cdrom", 5))
+	ret = hook (alias->name);
+      return ret;
     }
 
-  grub_devalias_iterate (dev_iterate);
-  return 0;
+  return grub_devalias_iterate (dev_iterate);
+}
+
+static char *
+compute_dev_path (const char *name)
+{
+  char *devpath = grub_malloc (grub_strlen (name) + 3);
+  char *p, c;
+
+  if (!devpath)
+    return NULL;
+
+  /* Un-escape commas. */
+  p = devpath;
+  while ((c = *name++) != '\0')
+    {
+      if (c == '\\' && *name == ',')
+	{
+	  *p++ = ',';
+	  name++;
+	}
+      else
+	*p++ = c;
+    }
+
+  if (! grub_ieee1275_test_flag (GRUB_IEEE1275_FLAG_NO_PARTITION_0))
+    {
+      *p++ = ':';
+      *p++ = '0';
+    }
+  *p++ = '\0';
+
+  return devpath;
 }
 
 static grub_err_t
@@ -76,68 +152,70 @@ grub_ofdisk_open (const char *name, grub_disk_t disk)
 {
   grub_ieee1275_phandle_t dev;
   grub_ieee1275_ihandle_t dev_ihandle = 0;
+  struct ofdisk_hash_ent *op;
   char *devpath;
   /* XXX: This should be large enough for any possible case.  */
   char prop[64];
   grub_ssize_t actual;
 
-  devpath = grub_strndup (name, grub_strlen (name) + 2);
+  devpath = compute_dev_path (name);
   if (! devpath)
     return grub_errno;
 
-  /* To access the complete disk add `:0'.  */
-  if (! grub_ieee1275_test_flag (GRUB_IEEE1275_FLAG_NO_PARTITION_0))
-    grub_strcat (devpath, ":0");
+  op = ofdisk_hash_find (devpath);
+  if (!op)
+    op = ofdisk_hash_add (devpath);
 
-  grub_dprintf ("disk", "Opening `%s'.\n", devpath);
+  grub_free (devpath);
+  if (!op)
+    return grub_errno;
 
-  grub_ieee1275_open (devpath, &dev_ihandle);
-  if (! dev_ihandle)
+  grub_dprintf ("disk", "Opening `%s'.\n", op->devpath);
+
+  if (grub_ieee1275_finddevice (op->devpath, &dev))
     {
-      grub_error (GRUB_ERR_UNKNOWN_DEVICE, "Can't open device");
-      goto fail;
-    }
-
-  grub_dprintf ("disk", "Opened `%s' as handle %p.\n", devpath, (void *) dev_ihandle);
-
-  if (grub_ieee1275_finddevice (devpath, &dev))
-    {
-      grub_error (GRUB_ERR_UNKNOWN_DEVICE, "Can't read device properties");
+      grub_error (GRUB_ERR_UNKNOWN_DEVICE, "can't read device properties");
       goto fail;
     }
 
   if (grub_ieee1275_get_property (dev, "device_type", prop, sizeof (prop),
 				  &actual))
     {
-      grub_error (GRUB_ERR_BAD_DEVICE, "Can't read the device type");
+      grub_error (GRUB_ERR_UNKNOWN_DEVICE, "can't read the device type");
       goto fail;
     }
 
   if (grub_strcmp (prop, "block"))
     {
-      grub_error (GRUB_ERR_BAD_DEVICE, "Not a block device");
+      grub_error (GRUB_ERR_BAD_DEVICE, "not a block device");
       goto fail;
     }
 
+  grub_ieee1275_open (op->devpath, &dev_ihandle);
+  if (! dev_ihandle)
+    {
+      grub_error (GRUB_ERR_UNKNOWN_DEVICE, "can't open device");
+      goto fail;
+    }
+
+  grub_dprintf ("disk", "Opened `%s' as handle %p.\n", op->devpath,
+		(void *) (unsigned long) dev_ihandle);
+
   /* XXX: There is no property to read the number of blocks.  There
      should be a property `#blocks', but it is not there.  Perhaps it
      is possible to use seek for this.  */
   disk->total_sectors = 0xFFFFFFFFUL;
 
-  /* XXX: Is it ok to use this?  Perhaps it is better to use the path
-     or some property.  */
-  disk->id = dev;
+  disk->id = (unsigned long) op;
 
   /* XXX: Read this, somehow.  */
   disk->has_partitions = 1;
-  disk->data = (void *) dev_ihandle;
-  grub_free (devpath);
+  disk->data = (void *) (unsigned long) dev_ihandle;
   return 0;
 
  fail:
   if (dev_ihandle)
     grub_ieee1275_close (dev_ihandle);
-  grub_free (devpath);
   return grub_errno;
 }
 
@@ -146,7 +224,7 @@ grub_ofdisk_close (grub_disk_t disk)
 {
   grub_dprintf ("disk", "Closing handle %p.\n",
 		(void *) disk->data);
-  grub_ieee1275_close ((grub_ieee1275_ihandle_t) disk->data);
+  grub_ieee1275_close ((grub_ieee1275_ihandle_t) (unsigned long) disk->data);
 }
 
 static grub_err_t
@@ -156,24 +234,20 @@ grub_ofdisk_read (grub_disk_t disk, grub_disk_addr_t sector,
   grub_ssize_t status, actual;
   unsigned long long pos;
 
-  grub_dprintf ("disk",
-		"Reading handle %p: sector 0x%llx, size 0x%lx, buf %p.\n",
-		(void *) disk->data, sector, (long) size, buf);
-
   pos = sector * 512UL;
 
-  grub_ieee1275_seek ((grub_ieee1275_ihandle_t) disk->data, (int) (pos >> 32),
-		      (int) pos & 0xFFFFFFFFUL, &status);
+  grub_ieee1275_seek ((grub_ieee1275_ihandle_t) (unsigned long) disk->data,
+		      pos, &status);
   if (status < 0)
     return grub_error (GRUB_ERR_READ_ERROR,
-		       "Seek error, can't seek block %llu",
-		       sector);
-  grub_ieee1275_read ((grub_ieee1275_ihandle_t) disk->data, buf,
-		      size * 512UL, &actual);
-  if (actual != actual)
-    return grub_error (GRUB_ERR_READ_ERROR, "Read error on block: %llu",
-		       sector);
-    
+		       "seek error, can't seek block %llu",
+		       (long long) sector);
+  grub_ieee1275_read ((grub_ieee1275_ihandle_t) (unsigned long) disk->data,
+		      buf, size * 512UL, &actual);
+  if (actual != (grub_ssize_t) (size * 512UL))
+    return grub_error (GRUB_ERR_READ_ERROR, "read error on block: %llu",
+		       (long long) sector);
+
   return 0;
 }
 
