@@ -1,7 +1,7 @@
 /* jfs.c - JFS.  */
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 2004,2005,2006,2007,2008  Free Software Foundation, Inc.
+ *  Copyright (C) 2004,2005,2006,2007,2008,2009  Free Software Foundation, Inc.
  *
  *  GRUB is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -24,6 +24,7 @@
 #include <grub/disk.h>
 #include <grub/dl.h>
 #include <grub/types.h>
+#include <grub/charset.h>
 
 #define GRUB_JFS_MAX_SYMLNK_CNT	8
 #define GRUB_JFS_FILETYPE_MASK	0170000
@@ -51,6 +52,8 @@ struct grub_jfs_sblock
 
   grub_uint8_t unused[71];
   grub_uint8_t volname[11];
+  grub_uint8_t unused2[32];
+  grub_uint8_t uuid[16];
 };
 
 struct grub_jfs_extent
@@ -341,7 +344,7 @@ grub_jfs_mount (grub_disk_t disk)
 
   if (grub_strncmp ((char *) (data->sblock.magic), "JFS1", 4))
     {
-      grub_error (GRUB_ERR_BAD_FS, "not a jfs filesystem");
+      grub_error (GRUB_ERR_BAD_FS, "not a JFS filesystem");
       goto fail;
     }
 
@@ -360,7 +363,7 @@ grub_jfs_mount (grub_disk_t disk)
   grub_free (data);
 
   if (grub_errno == GRUB_ERR_OUT_OF_RANGE)
-    grub_error (GRUB_ERR_BAD_FS, "not a jfs filesystem");
+    grub_error (GRUB_ERR_BAD_FS, "not a JFS filesystem");
 
   return 0;
 }
@@ -542,10 +545,6 @@ grub_jfs_read_file (struct grub_jfs_data *data,
   int i;
   int blockcnt;
 
-  /* Adjust len so it we can't read past the end of the file.  */
-  if (len > data->currinode.size)
-    len = data->currinode.size;
-
   blockcnt = ((len + pos + grub_le_to_cpu32 (data->sblock.blksz) - 1)
 	      / grub_le_to_cpu32 (data->sblock.blksz));
 
@@ -716,7 +715,7 @@ grub_jfs_lookup_symlink (struct grub_jfs_data *data, int ino)
 
   grub_jfs_find_file (data, symlink);
   if (grub_errno)
-    grub_error (grub_errno, "Can not follow symlink `%s'.", symlink);
+    grub_error (grub_errno, "cannot follow symlink `%s'", symlink);
 
   return grub_errno;
 }
@@ -832,6 +831,37 @@ grub_jfs_close (grub_file_t file)
   return GRUB_ERR_NONE;
 }
 
+static grub_err_t
+grub_jfs_uuid (grub_device_t device, char **uuid)
+{
+  struct grub_jfs_data *data;
+  grub_disk_t disk = device->disk;
+
+  grub_dl_ref (my_mod);
+
+  data = grub_jfs_mount (disk);
+  if (data)
+    {
+      *uuid = grub_xasprintf ("%02x%02x%02x%02x-%02x%02x-%02x%02x-%02x%02x-"
+			     "%02x%02x%02x%02x%02x%02x",
+			     data->sblock.uuid[0], data->sblock.uuid[1],
+			     data->sblock.uuid[2], data->sblock.uuid[3],
+			     data->sblock.uuid[4], data->sblock.uuid[5],
+			     data->sblock.uuid[6], data->sblock.uuid[7],
+			     data->sblock.uuid[8], data->sblock.uuid[9],
+			     data->sblock.uuid[10], data->sblock.uuid[11],
+			     data->sblock.uuid[12], data->sblock.uuid[13],
+			     data->sblock.uuid[14], data->sblock.uuid[15]);
+    }
+  else
+    *uuid = NULL;
+
+  grub_dl_unref (my_mod);
+
+  grub_free (data);
+
+  return grub_errno;
+}
 
 static grub_err_t
 grub_jfs_label (grub_device_t device, char **label)
@@ -856,6 +886,7 @@ static struct grub_fs grub_jfs_fs =
     .read = grub_jfs_read,
     .close = grub_jfs_close,
     .label = grub_jfs_label,
+    .uuid = grub_jfs_uuid,
     .next = 0
   };
 
