@@ -1,6 +1,6 @@
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 1999,2000,2001,2002,2003,2004,2007  Free Software Foundation, Inc.
+ *  Copyright (C) 1999,2000,2001,2002,2003,2004,2007,2008,2009  Free Software Foundation, Inc.
  *
  *  GRUB is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -21,7 +21,6 @@
 
 #define GRUB_LINUX_MAGIC_SIGNATURE	0x53726448      /* "HdrS" */
 #define GRUB_LINUX_DEFAULT_SETUP_SECTS	4
-#define GRUB_LINUX_FLAG_CAN_USE_HEAP	0x80
 #define GRUB_LINUX_INITRD_MAX_ADDRESS	0x37FFFFFF
 #define GRUB_LINUX_MAX_SETUP_SECTS	64
 #define GRUB_LINUX_BOOT_LOADER_TYPE	0x72
@@ -33,14 +32,15 @@
 #define GRUB_LINUX_SETUP_STACK		0x9000
 
 #define GRUB_LINUX_FLAG_BIG_KERNEL	0x1
+#define GRUB_LINUX_FLAG_QUIET		0x20
+#define GRUB_LINUX_FLAG_CAN_USE_HEAP	0x80
 
 /* Linux's video mode selection support. Actually I hate it!  */
 #define GRUB_LINUX_VID_MODE_NORMAL	0xFFFF
 #define GRUB_LINUX_VID_MODE_EXTENDED	0xFFFE
 #define GRUB_LINUX_VID_MODE_ASK		0xFFFD
+#define GRUB_LINUX_VID_MODE_VESA_START	0x0300
 
-#define GRUB_LINUX_CL_OFFSET		0x9000
-#define GRUB_LINUX_CL_END_OFFSET	0x90FF
 #define GRUB_LINUX_SETUP_MOVE_SIZE	0x9100
 #define GRUB_LINUX_CL_MAGIC		0xA33F
 
@@ -79,12 +79,13 @@ struct grub_e820_mmap
   grub_uint32_t type;
 } __attribute__((packed));
 
-#define GRUB_VIDEO_TYPE_VLFB	0x23    /* VESA VGA in graphic mode     */
-#define GRUB_VIDEO_TYPE_EFI	0x70
+#define GRUB_VIDEO_LINUX_TYPE_TEXT	0x01
+#define GRUB_VIDEO_LINUX_TYPE_VESA	0x23    /* VESA VGA in graphic mode.  */
+#define GRUB_VIDEO_LINUX_TYPE_SIMPLE	0x70    /* Linear framebuffer without any additional functions.  */
 
 /* For the Linux/i386 boot protocol version 2.03.  */
 struct linux_kernel_header
-{ 
+{
   grub_uint8_t code1[0x0020];
   grub_uint16_t cl_magic;		/* Magic number 0xA33F */
   grub_uint16_t cl_offset;		/* The offset of command line */
@@ -104,6 +105,17 @@ struct linux_kernel_header
   grub_uint16_t start_sys;		/* The load-low segment (obsolete) */
   grub_uint16_t kernel_version;		/* Points to kernel version string */
   grub_uint8_t type_of_loader;		/* Boot loader identifier */
+#define LINUX_LOADER_ID_LILO		0x0
+#define LINUX_LOADER_ID_LOADLIN		0x1
+#define LINUX_LOADER_ID_BOOTSECT	0x2
+#define LINUX_LOADER_ID_SYSLINUX	0x3
+#define LINUX_LOADER_ID_ETHERBOOT	0x4
+#define LINUX_LOADER_ID_ELILO		0x5
+#define LINUX_LOADER_ID_GRUB		0x7
+#define LINUX_LOADER_ID_UBOOT		0x8
+#define LINUX_LOADER_ID_XEN		0x9
+#define LINUX_LOADER_ID_GUJIN		0xa
+#define LINUX_LOADER_ID_QEMU		0xb
   grub_uint8_t loadflags;		/* Boot protocol option flags */
   grub_uint16_t setup_move_size;	/* Move to high memory size */
   grub_uint32_t code32_start;		/* Boot loader hook */
@@ -125,21 +137,21 @@ struct linux_kernel_params
   grub_uint8_t video_cursor_y;
 
   grub_uint16_t ext_mem;		/* 2 */
-  
+
   grub_uint16_t video_page;		/* 4 */
   grub_uint8_t video_mode;		/* 6 */
   grub_uint8_t video_width;		/* 7 */
-  
+
   grub_uint8_t padding1[0xa - 0x8];
-  
+
   grub_uint16_t video_ega_bx;		/* a */
-  
+
   grub_uint8_t padding2[0xe - 0xc];
-  
+
   grub_uint8_t video_height;		/* e */
   grub_uint8_t have_vga;		/* f */
   grub_uint16_t font_size;		/* 10 */
-  
+
   grub_uint16_t lfb_width;		/* 12 */
   grub_uint16_t lfb_height;		/* 14 */
   grub_uint16_t lfb_depth;		/* 16 */
@@ -162,8 +174,9 @@ struct linux_kernel_params
   grub_uint16_t vesapm_offset;		/* 30 */
   grub_uint16_t lfb_pages;		/* 32 */
   grub_uint16_t vesa_attrib;		/* 34 */
+  grub_uint32_t capabilities;		/* 36 */
 
-  grub_uint8_t padding3[0x40 - 0x36];
+  grub_uint8_t padding3[0x40 - 0x3a];
 
   grub_uint16_t apm_version;		/* 40 */
   grub_uint16_t apm_code_segment;	/* 42 */
@@ -173,16 +186,16 @@ struct linux_kernel_params
   grub_uint16_t apm_flags;		/* 4c */
   grub_uint32_t apm_code_len;		/* 4e */
   grub_uint16_t apm_data_len;		/* 52 */
-  
+
   grub_uint8_t padding4[0x60 - 0x54];
-  
+
   grub_uint32_t ist_signature;		/* 60 */
   grub_uint32_t ist_command;		/* 64 */
   grub_uint32_t ist_event;		/* 68 */
   grub_uint32_t ist_perf_level;		/* 6c */
 
   grub_uint8_t padding5[0x80 - 0x70];
-  
+
   grub_uint8_t hd0_drive_info[0x10];	/* 80 */
   grub_uint8_t hd1_drive_info[0x10];	/* 90 */
   grub_uint16_t rom_config_len;		/* a0 */
@@ -222,15 +235,15 @@ struct linux_kernel_params
           grub_uint32_t efi_mmap_hi;		/* 1dc */
         } v0206;
     };
-  
-  grub_uint32_t alt_mem;		/* 1e0 */
-  
-  grub_uint8_t padding8[0x1e8 - 0x1e4];
-  
-  grub_uint32_t mmap_size;		/* 1e8 */
 
-  grub_uint8_t padding9[0x1f1 - 0x1ec];
-  
+  grub_uint32_t alt_mem;		/* 1e0 */
+
+  grub_uint8_t padding8[0x1e8 - 0x1e4];
+
+  grub_uint8_t mmap_size;		/* 1e8 */
+
+  grub_uint8_t padding9[0x1f1 - 0x1e9];
+
   grub_uint8_t setup_sects;		/* The size of the setup in sectors */
   grub_uint16_t root_flags;		/* If the root is mounted readonly */
   grub_uint16_t syssize;		/* obsolete */
