@@ -30,19 +30,19 @@ struct grub_partition_map
 {
   /* The name of the partition map type.  */
   const char *name;
-  
+
   /* Call HOOK with each partition, until HOOK returns non-zero.  */
   grub_err_t (*iterate) (struct grub_disk *disk,
 			 int (*hook) (struct grub_disk *disk,
 				      const grub_partition_t partition));
-  
+
   /* Return the partition named STR on the disk DISK.  */
   grub_partition_t (*probe) (struct grub_disk *disk,
 			     const char *str);
-  
+
   /* Return the name of the partition PARTITION.  */
   char *(*get_name) (const grub_partition_t partition);
-  
+
   /* The next partition map type.  */
   struct grub_partition_map *next;
 };
@@ -62,10 +62,10 @@ struct grub_partition
 
   /* The index of this partition in the partition table.  */
   int index;
-  
+
   /* Partition map type specific data.  */
   void *data;
-  
+
   /* The type partition map.  */
   grub_partition_map_t partmap;
 };
@@ -78,25 +78,11 @@ int EXPORT_FUNC(grub_partition_iterate) (struct grub_disk *disk,
 char *EXPORT_FUNC(grub_partition_get_name) (const grub_partition_t partition);
 
 int EXPORT_FUNC(grub_partition_map_iterate) (int (*hook) (const grub_partition_map_t partmap));
-					      
+
 void EXPORT_FUNC(grub_partition_map_register) (grub_partition_map_t partmap);
 
 void EXPORT_FUNC(grub_partition_map_unregister) (grub_partition_map_t partmap);
 
-#ifdef GRUB_UTIL
-void grub_pc_partition_map_init (void);
-void grub_pc_partition_map_fini (void);
-void grub_amiga_partition_map_init (void);
-void grub_amiga_partition_map_fini (void);
-void grub_apple_partition_map_init (void);
-void grub_apple_partition_map_fini (void);
-void grub_sun_partition_map_init (void);
-void grub_sun_partition_map_fini (void);
-void grub_gpt_partition_map_init (void);
-void grub_gpt_partition_map_fini (void);
-void grub_apple_partition_map_init (void);
-void grub_apple_partition_map_fini (void);
-#endif
 
 static inline grub_disk_addr_t
 grub_partition_get_start (const grub_partition_t p)
