@@ -19,19 +19,6 @@
 #ifndef GRUB_CONSOLE_MACHINE_HEADER
 #define GRUB_CONSOLE_MACHINE_HEADER	1
 
-/* Define scan codes.  */
-#define GRUB_CONSOLE_KEY_LEFT		0x4B00
-#define GRUB_CONSOLE_KEY_RIGHT		0x4D00
-#define GRUB_CONSOLE_KEY_UP		0x4800
-#define GRUB_CONSOLE_KEY_DOWN		0x5000
-#define GRUB_CONSOLE_KEY_IC		0x5200
-#define GRUB_CONSOLE_KEY_DC		0x5300
-#define GRUB_CONSOLE_KEY_BACKSPACE	0x0008
-#define GRUB_CONSOLE_KEY_HOME		0x4700
-#define GRUB_CONSOLE_KEY_END		0x4F00
-#define GRUB_CONSOLE_KEY_NPAGE		0x4900
-#define GRUB_CONSOLE_KEY_PPAGE		0x5100
-
 /* Initialize the console system.  */
 void grub_console_init (void);
 
