@@ -1,7 +1,7 @@
 /* elf.c - load ELF files */
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 2003,2004,2005,2006,2007,2008  Free Software Foundation, Inc.
+ *  Copyright (C) 2003,2004,2005,2006,2007,2008,2009  Free Software Foundation, Inc.
  *
  *  GRUB is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -61,21 +61,20 @@ grub_elf_file (grub_file_t file)
 {
   grub_elf_t elf;
 
-  elf = grub_malloc (sizeof (*elf));
+  elf = grub_zalloc (sizeof (*elf));
   if (! elf)
     return 0;
 
   elf->file = file;
-  elf->phdrs = 0;
 
   if (grub_file_seek (elf->file, 0) == (grub_off_t) -1)
     goto fail;
 
-  if (grub_file_read (elf->file, (char *) &elf->ehdr, sizeof (elf->ehdr))
+  if (grub_file_read (elf->file, &elf->ehdr, sizeof (elf->ehdr))
       != sizeof (elf->ehdr))
     {
       grub_error_push ();
-      grub_error (GRUB_ERR_READ_ERROR, "Cannot read ELF header.");
+      grub_error (GRUB_ERR_READ_ERROR, "cannot read ELF header");
       goto fail;
     }
 
@@ -123,9 +122,9 @@ grub_elf32_load_phdrs (grub_elf_t elf)
 
   phdrs_size = elf->ehdr.ehdr32.e_phnum * elf->ehdr.ehdr32.e_phentsize;
 
-  grub_dprintf ("elf", "Loading program headers at 0x%llx, size 0x%x.\n",
+  grub_dprintf ("elf", "Loading program headers at 0x%llx, size 0x%lx.\n",
 		(unsigned long long) elf->ehdr.ehdr32.e_phoff,
-		phdrs_size);
+		(unsigned long) phdrs_size);
 
   elf->phdrs = grub_malloc (phdrs_size);
   if (! elf->phdrs)
@@ -135,7 +134,7 @@ grub_elf32_load_phdrs (grub_elf_t elf)
       || (grub_file_read (elf->file, elf->phdrs, phdrs_size) != phdrs_size))
     {
       grub_error_push ();
-      return grub_error (GRUB_ERR_READ_ERROR, "Cannot read program headers");
+      return grub_error (GRUB_ERR_READ_ERROR, "cannot read program headers");
     }
 
   return GRUB_ERR_NONE;
@@ -173,7 +172,7 @@ grub_elf32_phdr_iterate (grub_elf_t elf,
 
 /* Calculate the amount of memory spanned by the segments.  */
 grub_size_t
-grub_elf32_size (grub_elf_t elf)
+grub_elf32_size (grub_elf_t elf, Elf32_Addr *base)
 {
   Elf32_Addr segments_start = (Elf32_Addr) -1;
   Elf32_Addr segments_end = 0;
@@ -182,7 +181,9 @@ grub_elf32_size (grub_elf_t elf)
   /* Run through the program headers to calculate the total memory size we
    * should claim.  */
   auto int NESTED_FUNC_ATTR calcsize (grub_elf_t _elf, Elf32_Phdr *phdr, void *_arg);
-  int NESTED_FUNC_ATTR calcsize (grub_elf_t UNUSED _elf, Elf32_Phdr *phdr, void UNUSED *_arg)
+  int NESTED_FUNC_ATTR calcsize (grub_elf_t _elf  __attribute__ ((unused)),
+				 Elf32_Phdr *phdr,
+				 void *_arg __attribute__ ((unused)))
     {
       /* Only consider loadable segments.  */
       if (phdr->p_type != PT_LOAD)
@@ -197,23 +198,28 @@ grub_elf32_size (grub_elf_t elf)
 
   grub_elf32_phdr_iterate (elf, calcsize, 0);
 
+  if (base)
+    *base = 0;
+
   if (nr_phdrs == 0)
     {
-      grub_error (GRUB_ERR_BAD_OS, "No program headers present");
+      grub_error (GRUB_ERR_BAD_OS, "no program headers present");
       return 0;
     }
 
   if (segments_end < segments_start)
     {
       /* Very bad addresses.  */
-      grub_error (GRUB_ERR_BAD_OS, "Bad program header load addresses");
+      grub_error (GRUB_ERR_BAD_OS, "bad program header load addresses");
       return 0;
     }
 
+  if (base)
+    *base = segments_start;
+
   return segments_end - segments_start;
 }
 
-
 /* Load every loadable segment into memory specified by `_load_hook'.  */
 grub_err_t
 grub_elf32_load (grub_elf_t _elf, grub_elf32_load_hook_t _load_hook,
@@ -228,13 +234,14 @@ grub_elf32_load (grub_elf_t _elf, grub_elf32_load_hook_t _load_hook,
   {
     grub_elf32_load_hook_t load_hook = (grub_elf32_load_hook_t) hook;
     grub_addr_t load_addr;
+    int do_load = 1;
 
-    if (phdr->p_type != PT_LOAD)
-      return 0;
-
-    if (load_hook && load_hook (phdr, &load_addr))
-      return 1;
     load_addr = phdr->p_paddr;
+    if (load_hook && load_hook (phdr, &load_addr, &do_load))
+      return 1;
+
+    if (! do_load)
+      return 0;
 
     if (load_addr < load_base)
       load_base = load_addr;
@@ -247,7 +254,7 @@ grub_elf32_load (grub_elf_t _elf, grub_elf32_load_hook_t _load_hook,
       {
 	grub_error_push ();
 	return grub_error (GRUB_ERR_BAD_OS,
-			   "Invalid offset in program header.");
+			   "invalid offset in program header");
       }
 
     if (phdr->p_filesz)
@@ -259,8 +266,8 @@ grub_elf32_load (grub_elf_t _elf, grub_elf32_load_hook_t _load_hook,
 	    /* XXX How can we free memory from `load_hook'? */
 	    grub_error_push ();
 	    return grub_error (GRUB_ERR_BAD_OS,
-			       "Couldn't read segment from file: "
-			       "wanted 0x%lx bytes; read 0x%lx bytes.",
+			       "couldn't read segment from file: "
+			       "wanted 0x%lx bytes; read 0x%lx bytes",
 			       phdr->p_filesz, read);
 	  }
       }
@@ -301,9 +308,9 @@ grub_elf64_load_phdrs (grub_elf_t elf)
 
   phdrs_size = elf->ehdr.ehdr64.e_phnum * elf->ehdr.ehdr64.e_phentsize;
 
-  grub_dprintf ("elf", "Loading program headers at 0x%llx, size 0x%x.\n",
+  grub_dprintf ("elf", "Loading program headers at 0x%llx, size 0x%lx.\n",
 		(unsigned long long) elf->ehdr.ehdr64.e_phoff,
-		phdrs_size);
+		(unsigned long) phdrs_size);
 
   elf->phdrs = grub_malloc (phdrs_size);
   if (! elf->phdrs)
@@ -313,7 +320,7 @@ grub_elf64_load_phdrs (grub_elf_t elf)
       || (grub_file_read (elf->file, elf->phdrs, phdrs_size) != phdrs_size))
     {
       grub_error_push ();
-      return grub_error (GRUB_ERR_READ_ERROR, "Cannot read program headers");
+      return grub_error (GRUB_ERR_READ_ERROR, "cannot read program headers");
     }
 
   return GRUB_ERR_NONE;
@@ -351,7 +358,7 @@ grub_elf64_phdr_iterate (grub_elf_t elf,
 
 /* Calculate the amount of memory spanned by the segments.  */
 grub_size_t
-grub_elf64_size (grub_elf_t elf)
+grub_elf64_size (grub_elf_t elf, Elf64_Addr *base)
 {
   Elf64_Addr segments_start = (Elf64_Addr) -1;
   Elf64_Addr segments_end = 0;
@@ -360,7 +367,9 @@ grub_elf64_size (grub_elf_t elf)
   /* Run through the program headers to calculate the total memory size we
    * should claim.  */
   auto int NESTED_FUNC_ATTR calcsize (grub_elf_t _elf, Elf64_Phdr *phdr, void *_arg);
-  int NESTED_FUNC_ATTR calcsize (grub_elf_t UNUSED _elf, Elf64_Phdr *phdr, void UNUSED *_arg)
+  int NESTED_FUNC_ATTR calcsize (grub_elf_t _elf __attribute__ ((unused)),
+				 Elf64_Phdr *phdr,
+				 void *_arg __attribute__ ((unused)))
     {
       /* Only consider loadable segments.  */
       if (phdr->p_type != PT_LOAD)
@@ -375,19 +384,25 @@ grub_elf64_size (grub_elf_t elf)
 
   grub_elf64_phdr_iterate (elf, calcsize, 0);
 
+  if (base)
+    *base = 0;
+
   if (nr_phdrs == 0)
     {
-      grub_error (GRUB_ERR_BAD_OS, "No program headers present");
+      grub_error (GRUB_ERR_BAD_OS, "no program headers present");
       return 0;
     }
 
   if (segments_end < segments_start)
     {
       /* Very bad addresses.  */
-      grub_error (GRUB_ERR_BAD_OS, "Bad program header load addresses");
+      grub_error (GRUB_ERR_BAD_OS, "bad program header load addresses");
       return 0;
     }
 
+  if (base)
+    *base = segments_start;
+
   return segments_end - segments_start;
 }
 
@@ -407,13 +422,14 @@ grub_elf64_load (grub_elf_t _elf, grub_elf64_load_hook_t _load_hook,
   {
     grub_elf64_load_hook_t load_hook = (grub_elf64_load_hook_t) hook;
     grub_addr_t load_addr;
+    int do_load = 1;
 
-    if (phdr->p_type != PT_LOAD)
-      return 0;
-
-    if (load_hook && load_hook (phdr, &load_addr))
-      return 1;
     load_addr = phdr->p_paddr;
+    if (load_hook && load_hook (phdr, &load_addr, &do_load))
+      return 1;
+
+    if (! do_load)
+      return 0;
 
     if (load_addr < load_base)
       load_base = load_addr;
@@ -426,7 +442,7 @@ grub_elf64_load (grub_elf_t _elf, grub_elf64_load_hook_t _load_hook,
       {
 	grub_error_push ();
 	return grub_error (GRUB_ERR_BAD_OS,
-			   "Invalid offset in program header.");
+			   "invalid offset in program header");
       }
 
     if (phdr->p_filesz)
@@ -438,8 +454,8 @@ grub_elf64_load (grub_elf_t _elf, grub_elf64_load_hook_t _load_hook,
 	    /* XXX How can we free memory from `load_hook'?  */
 	    grub_error_push ();
 	    return grub_error (GRUB_ERR_BAD_OS,
-			      "Couldn't read segment from file: "
-			      "wanted 0x%lx bytes; read 0x%lx bytes.",
+			      "couldn't read segment from file: "
+			      "wanted 0x%lx bytes; read 0x%lx bytes",
 			      phdr->p_filesz, read);
           }
       }
