@@ -18,14 +18,11 @@
  */
 
 #include <grub/types.h>
+#include <grub/cache.h>
+
+void grub_stop_floppy (void);
 
 void
 grub_stop_floppy (void)
 {
 }
-
-void 
-grub_arch_sync_caches (void *address __attribute__ ((unused)),
-		       grub_size_t len __attribute__ ((unused)))
-{
-}
