@@ -1,7 +1,7 @@
 /* misc.c - definitions of misc functions */
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 1999,2000,2001,2002,2003,2004,2005,2006,2007,2008  Free Software Foundation, Inc.
+ *  Copyright (C) 1999,2000,2001,2002,2003,2004,2005,2006,2007,2008,2009,2010  Free Software Foundation, Inc.
  *
  *  GRUB is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -23,7 +23,25 @@
 #include <stdarg.h>
 #include <grub/term.h>
 #include <grub/env.h>
-#include <grub/time.h>
+#include <grub/i18n.h>
+
+static int
+grub_vsnprintf_real (char *str, grub_size_t n, const char *fmt, va_list args);
+
+static int
+grub_iswordseparator (int c)
+{
+  return (grub_isspace (c) || c == ',' || c == ';' || c == '|' || c == '&');
+}
+
+/* grub_gettext_dummy is not translating anything.  */
+static const char *
+grub_gettext_dummy (const char *s)
+{
+  return s;
+}
+
+const char* (*grub_gettext) (const char *s) = grub_gettext_dummy;
 
 void *
 grub_memmove (void *dest, const void *src, grub_size_t n)
@@ -38,18 +56,30 @@ grub_memmove (void *dest, const void *src, grub_size_t n)
     {
       d += n;
       s += n;
-      
+
       while (n--)
 	*--d = *--s;
     }
-  
+
   return dest;
 }
+
+#ifndef APPLE_CC
 void *memmove (void *dest, const void *src, grub_size_t n)
   __attribute__ ((alias ("grub_memmove")));
 /* GCC emits references to memcpy() for struct copies etc.  */
 void *memcpy (void *dest, const void *src, grub_size_t n)
   __attribute__ ((alias ("grub_memmove")));
+#else
+void *memcpy (void *dest, const void *src, grub_size_t n)
+{
+	return grub_memmove (dest, src, n);
+}
+void *memmove (void *dest, const void *src, grub_size_t n)
+{
+	return grub_memmove (dest, src, n);
+}
+#endif
 
 char *
 grub_strcpy (char *dest, const char *src)
@@ -66,7 +96,7 @@ char *
 grub_strncpy (char *dest, const char *src, int c)
 {
   char *p = dest;
-  
+
   while ((*p++ = *src++) != '\0' && --c)
     ;
 
@@ -86,54 +116,70 @@ grub_stpcpy (char *dest, const char *src)
   return d - 1;
 }
 
-char *
-grub_strcat (char *dest, const char *src)
-{
-  char *p = dest;
-
-  while (*p)
-    p++;
-
-  while ((*p = *src) != '\0')
-    {
-      p++;
-      src++;
-    }
-
-  return dest;
-}
-
-char *
-grub_strncat (char *dest, const char *src, int c)
-{
-  char *p = dest;
-
-  while (*p)
-    p++;
-
-  while ((*p = *src) != '\0' && c--)
-    {
-      p++;
-      src++;
-    }
-
-  *p = '\0';
-
-  return dest;
-}
-
 int
 grub_printf (const char *fmt, ...)
 {
   va_list ap;
   int ret;
-  
+
   va_start (ap, fmt);
   ret = grub_vprintf (fmt, ap);
   va_end (ap);
 
   return ret;
-}  
+}
+
+int
+grub_printf_ (const char *fmt, ...)
+{
+  va_list ap;
+  int ret;
+
+  va_start (ap, fmt);
+  ret = grub_vprintf (_(fmt), ap);
+  va_end (ap);
+
+  return ret;
+}
+
+int
+grub_puts (const char *s)
+{
+  while (*s)
+    {
+      grub_putchar (*s);
+      s++;
+    }
+  grub_putchar ('\n');
+
+  return 1;	/* Cannot fail.  */
+}
+
+int
+grub_puts_ (const char *s)
+{
+  return grub_puts (_(s));
+}
+
+#if defined (APPLE_CC) && ! defined (GRUB_UTIL)
+int
+grub_err_printf (const char *fmt, ...)
+{
+	va_list ap;
+	int ret;
+
+	va_start (ap, fmt);
+	ret = grub_vprintf (fmt, ap);
+	va_end (ap);
+
+	return ret;
+}
+#endif
+
+#if ! defined (APPLE_CC) && ! defined (GRUB_UTIL)
+int grub_err_printf (const char *fmt, ...)
+__attribute__ ((alias("grub_printf")));
+#endif
 
 void
 grub_real_dprintf (const char *file, const int line, const char *condition,
@@ -141,10 +187,10 @@ grub_real_dprintf (const char *file, const int line, const char *condition,
 {
   va_list args;
   const char *debug = grub_env_get ("debug");
-  
+
   if (! debug)
     return;
-  
+
   if (grub_strword (debug, "all") || grub_strword (debug, condition))
     {
       grub_printf ("%s:%d: ", file, line);
@@ -159,8 +205,7 @@ grub_vprintf (const char *fmt, va_list args)
 {
   int ret;
 
-  ret = grub_vsprintf (0, fmt, args);
-  grub_refresh ();
+  ret = grub_vsnprintf_real (0, 0, fmt, args);
   return ret;
 }
 
@@ -169,7 +214,7 @@ grub_memcmp (const void *s1, const void *s2, grub_size_t n)
 {
   const char *t1 = s1;
   const char *t2 = s2;
-  
+
   while (n--)
     {
       if (*t1 != *t2)
@@ -181,8 +226,15 @@ grub_memcmp (const void *s1, const void *s2, grub_size_t n)
 
   return 0;
 }
+#ifndef APPLE_CC
 int memcmp (const void *s1, const void *s2, grub_size_t n)
   __attribute__ ((alias ("grub_memcmp")));
+#else
+int memcmp (const void *s1, const void *s2, grub_size_t n)
+{
+  return grub_memcmp (s1, s2, n);
+}
+#endif
 
 int
 grub_strcmp (const char *s1, const char *s2)
@@ -190,8 +242,8 @@ grub_strcmp (const char *s1, const char *s2)
   while (*s1 && *s2)
     {
       if (*s1 != *s2)
-	return (int) *s1 - (int) *s2;
-      
+	break;
+
       s1++;
       s2++;
     }
@@ -204,12 +256,12 @@ grub_strncmp (const char *s1, const char *s2, grub_size_t n)
 {
   if (n == 0)
     return 0;
-  
+
   while (*s1 && *s2 && --n)
     {
       if (*s1 != *s2)
-	return (int) *s1 - (int) *s2;
-      
+	break;
+
       s1++;
       s2++;
     }
@@ -217,33 +269,15 @@ grub_strncmp (const char *s1, const char *s2, grub_size_t n)
   return (int) *s1 - (int) *s2;
 }
 
-int
-grub_strncasecmp (const char *s1, const char *s2, int c)
-{
-  int p = 1;
-
-  while (grub_tolower (*s1) && grub_tolower (*s2) && p < c)
-    {
-      if (grub_tolower (*s1) != grub_tolower (*s2))
-	return (int) grub_tolower (*s1) - (int) grub_tolower (*s2);
-      
-      s1++;
-      s2++;
-      p++;
-    }
-
-  return (int) *s1 - (int) *s2;
-}
-
 char *
 grub_strchr (const char *s, int c)
 {
-  while (*s)
+  do
     {
       if (*s == c)
 	return (char *) s;
-      s++;
     }
+  while (*s++);
 
   return 0;
 }
@@ -251,14 +285,14 @@ grub_strchr (const char *s, int c)
 char *
 grub_strrchr (const char *s, int c)
 {
-  char *p = 0;
+  char *p = NULL;
 
-  while (*s)
+  do
     {
       if (*s == c)
 	p = (char *) s;
-      s++;
     }
+  while (*s++);
 
   return p;
 }
@@ -278,7 +312,7 @@ grub_strstr (const char *haystack, const char *needle)
       /* Speed up the following searches of needle by caching its first
 	 character.  */
       char b = *needle++;
-      
+
       for (;; haystack++)
 	{
 	  if (*haystack == '\0')
@@ -346,12 +380,6 @@ grub_strword (const char *haystack, const char *needle)
   return 0;
 }
 
-int
-grub_iswordseparator (int c)
-{
-  return (grub_isspace (c) || c == ',' || c == ';' || c == '|' || c == '&');
-}
-
 int
 grub_isspace (int c)
 {
@@ -364,33 +392,6 @@ grub_isprint (int c)
   return (c >= ' ' && c <= '~');
 }
 
-int
-grub_isalpha (int c)
-{
-  return (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z');
-}
-
-int
-grub_isdigit (int c)
-{
-  return (c >= '0' && c <= '9');
-}
-
-int
-grub_isgraph (int c)
-{
-  return (c >= '!' && c <= '~');
-}
-
-int
-grub_tolower (int c)
-{
-  if (c >= 'A' && c <= 'Z')
-    return c - 'A' + 'a';
-
-  return c;
-}
-
 
 unsigned long
 grub_strtoul (const char *str, char **end, int base)
@@ -412,14 +413,14 @@ grub_strtoull (const char *str, char **end, int base)
 {
   unsigned long long num = 0;
   int found = 0;
-  
+
   /* Skip white spaces.  */
   while (*str && grub_isspace (*str))
     str++;
-  
+
   /* Guess the base, if not specified. The prefix `0x' means 16, and
      the prefix `0' means 8.  */
-  if (base == 0 && str[0] == '0')
+  if (str[0] == '0')
     {
       if (str[1] == 'x')
 	{
@@ -429,10 +430,10 @@ grub_strtoull (const char *str, char **end, int base)
 	      str += 2;
 	    }
 	}
-      else if (str[1] >= '0' && str[1] <= '7')
+      else if (base == 0 && str[1] >= '0' && str[1] <= '7')
 	base = 8;
     }
-  
+
   if (base == 0)
     base = 10;
 
@@ -466,7 +467,7 @@ grub_strtoull (const char *str, char **end, int base)
       grub_error (GRUB_ERR_BAD_NUMBER, "unrecognized number");
       return 0;
     }
-  
+
   if (end)
     *end = (char *) str;
 
@@ -478,7 +479,7 @@ grub_strdup (const char *s)
 {
   grub_size_t len;
   char *p;
-  
+
   len = grub_strlen (s) + 1;
   p = (char *) grub_malloc (len);
   if (! p)
@@ -492,14 +493,14 @@ grub_strndup (const char *s, grub_size_t n)
 {
   grub_size_t len;
   char *p;
-  
+
   len = grub_strlen (s);
   if (len > n)
     len = n;
   p = (char *) grub_malloc (len + 1);
   if (! p)
     return 0;
-  
+
   grub_memcpy (p, s, len);
   p[len] = '\0';
   return p;
@@ -515,8 +516,15 @@ grub_memset (void *s, int c, grub_size_t n)
 
   return s;
 }
+#ifndef APPLE_CC
 void *memset (void *s, int c, grub_size_t n)
   __attribute__ ((alias ("grub_memset")));
+#else
+void *memset (void *s, int c, grub_size_t n)
+{
+  return grub_memset (s, c, n);
+}
+#endif
 
 grub_size_t
 grub_strlen (const char *s)
@@ -546,31 +554,6 @@ grub_reverse (char *str)
     }
 }
 
-static char *
-grub_itoa (char *str, int c, unsigned n)
-{
-  unsigned base = (c == 'x') ? 16 : 10;
-  char *p;
-  
-  if ((int) n < 0 && c == 'd')
-    {
-      n = (unsigned) (-((int) n));
-      *str++ = '-';
-    }
-
-  p = str;
-  do
-    {
-      unsigned d = n % base;
-      *p++ = (d > 9) ? d + 'a' - 10 : d + '0';
-    }
-  while (n /= base);
-  *p = 0;
-
-  grub_reverse (str);
-  return p;
-}
-
 /* Divide N by D, return the quotient, and store the remainder in *R.  */
 grub_uint64_t
 grub_divmod64 (grub_uint64_t n, grub_uint32_t d, grub_uint32_t *r)
@@ -591,17 +574,17 @@ grub_divmod64 (grub_uint64_t n, grub_uint32_t d, grub_uint32_t *r)
 
       return ((grub_uint32_t) n) / d;
     }
-  
+
   while (bits--)
     {
       m <<= 1;
-      
+
       if (n & (1ULL << 63))
 	m |= 1;
-      
+
       q <<= 1;
       n <<= 1;
-      
+
       if (m >= d)
 	{
 	  q |= 1;
@@ -611,7 +594,7 @@ grub_divmod64 (grub_uint64_t n, grub_uint32_t d, grub_uint32_t *r)
 
   if (r)
     *r = m;
-  
+
   return q;
 }
 
@@ -622,7 +605,7 @@ grub_lltoa (char *str, int c, unsigned long long n)
 {
   unsigned base = (c == 'x') ? 16 : 10;
   char *p;
-  
+
   if ((long long) n < 0 && c == 'd')
     {
       n = (unsigned long long) (-((long long) n));
@@ -643,49 +626,34 @@ grub_lltoa (char *str, int c, unsigned long long n)
     do
       {
 	unsigned m;
-	
+
 	n = grub_divmod64 (n, 10, &m);
 	*p++ = m + '0';
       }
     while (n);
-  
+
   *p = 0;
 
   grub_reverse (str);
   return p;
 }
 
-static char *
-grub_ftoa (char *str, double f, int round)
-{
-  unsigned int intp;
-  unsigned int fractp;
-  unsigned int power = 1;
-  int i;
-
-  for (i = 0; i < round; i++)
-    power *= 10;
-
-  intp = f;
-  fractp = (f - (float) intp) * power;
-
-  grub_sprintf (str, "%d.%d", intp, fractp);
-  return str;
-}
-
-int
-grub_vsprintf (char *str, const char *fmt, va_list args)
+static int
+grub_vsnprintf_real (char *str, grub_size_t max_len, const char *fmt, va_list args)
 {
   char c;
-  int count = 0;
+  grub_size_t count = 0;
   auto void write_char (unsigned char ch);
   auto void write_str (const char *s);
   auto void write_fill (const char ch, int n);
-  
+
   void write_char (unsigned char ch)
     {
       if (str)
-	*str++ = ch;
+	{
+	  if (count < max_len)
+	    *str++ = ch;
+	}
       else
 	grub_putchar (ch);
 
@@ -704,7 +672,7 @@ grub_vsprintf (char *str, const char *fmt, va_list args)
       for (i = 0; i < n; i++)
 	write_char (ch);
     }
-  
+
   while ((c = *fmt++) != 0)
     {
       if (c != '%')
@@ -714,12 +682,13 @@ grub_vsprintf (char *str, const char *fmt, va_list args)
 	  char tmp[32];
 	  char *p;
 	  unsigned int format1 = 0;
-	  unsigned int format2 = 3;
+	  unsigned int format2 = ~ 0U;
 	  char zerofill = ' ';
 	  int rightfill = 0;
 	  int n;
 	  int longfmt = 0;
 	  int longlongfmt = 0;
+	  int unsig = 0;
 
 	  if (*fmt && *fmt =='-')
 	    {
@@ -741,20 +710,22 @@ grub_vsprintf (char *str, const char *fmt, va_list args)
 		zerofill = '0';
 	      format1 = grub_strtoul (s, 0, 10);
 	      fmt = p;
-	      if (*p && *p == '.')
+	    }
+
+	  if (*p && *p == '.')
+	    {
+	      p++;
+	      fmt++;
+	      while (*p && grub_isdigit (*p))
+		p++;
+
+	      if (p > fmt)
 		{
-		  p++;
-		  fmt++;
-		  while (*p && grub_isdigit (*p))
-		    p++;
-		  
-		  if (p > fmt)
-		    {
-		      char fstr[p - fmt];
-		      grub_strncpy (fstr, fmt, p - fmt);
-		      format2 = grub_strtoul (fstr, 0, 10);
-		      fmt = p;
-		    }
+		  char fstr[p - fmt + 1];
+		  grub_strncpy (fstr, fmt, p - fmt);
+		  fstr[p - fmt] = 0;
+		  format2 = grub_strtoul (fstr, 0, 10);
+		  fmt = p;
 		}
 	    }
 
@@ -776,9 +747,11 @@ grub_vsprintf (char *str, const char *fmt, va_list args)
 	      write_str ("0x");
 	      c = 'x';
 	      longlongfmt |= (sizeof (void *) == sizeof (long long));
-	      /* fall through */
+	      /* Fall through. */
 	    case 'x':
 	    case 'u':
+	      unsig = 1;
+	      /* Fall through. */
 	    case 'd':
 	      if (longlongfmt)
 		{
@@ -787,13 +760,25 @@ grub_vsprintf (char *str, const char *fmt, va_list args)
 		  ll = va_arg (args, long long);
 		  grub_lltoa (tmp, c, ll);
 		}
+	      else if (longfmt && unsig)
+		{
+		  unsigned long l = va_arg (args, unsigned long);
+		  grub_lltoa (tmp, c, l);
+		}
+	      else if (longfmt)
+		{
+		  long l = va_arg (args, long);
+		  grub_lltoa (tmp, c, l);
+		}
+	      else if (unsig)
+		{
+		  unsigned u = va_arg (args, unsigned);
+		  grub_lltoa (tmp, c, u);
+		}
 	      else
 		{
-		  if (longfmt)
-		    n = va_arg (args, long);
-		  else
-		    n = va_arg (args, int);
-		  grub_itoa (tmp, c, n);
+		  n = va_arg (args, int);
+		  grub_lltoa (tmp, c, n);
 		}
 	      if (! rightfill && grub_strlen (tmp) < format1)
 		write_fill (zerofill, format1 - grub_strlen (tmp));
@@ -801,31 +786,18 @@ grub_vsprintf (char *str, const char *fmt, va_list args)
 	      if (rightfill && grub_strlen (tmp) < format1)
 		write_fill (zerofill, format1 - grub_strlen (tmp));
 	      break;
-	      
+
 	    case 'c':
 	      n = va_arg (args, int);
 	      write_char (n & 0xff);
 	      break;
 
-	    case 'f':
-	      {
-		float f;
-		f = va_arg (args, double);
-		grub_ftoa (tmp, f, format2);
-		if (!rightfill && grub_strlen (tmp) < format1)
-		  write_fill (zerofill, format1 - grub_strlen (tmp));
-		write_str (tmp);
-		if (rightfill && grub_strlen (tmp) < format1)
-		  write_fill (zerofill, format1 - grub_strlen (tmp));
-		break;
-	      }
-	      
 	    case 'C':
 	      {
 		grub_uint32_t code = va_arg (args, grub_uint32_t);
 		int shift;
 		unsigned mask;
-		
+
 		if (code <= 0x7f)
 		  {
 		    shift = 0;
@@ -864,7 +836,7 @@ grub_vsprintf (char *str, const char *fmt, va_list args)
 		  }
 
 		write_char (mask | (code >> shift));
-		
+
 		for (shift -= 6; shift >= 0; shift -= 6)
 		  write_char (0x80 | (0x3f & (code >> shift)));
 	      }
@@ -874,17 +846,23 @@ grub_vsprintf (char *str, const char *fmt, va_list args)
 	      p = va_arg (args, char *);
 	      if (p)
 		{
-		  if (!rightfill && grub_strlen (p) < format1)
-		    write_fill (zerofill, format1 - grub_strlen (p));
-		  
-		  write_str (p);
-		  
-		  if (rightfill && grub_strlen (p) < format1)
-		    write_fill (zerofill, format1 - grub_strlen (p));
+		  grub_size_t len = 0;
+		  while (len < format2 && p[len])
+		    len++;
+
+		  if (!rightfill && len < format1)
+		    write_fill (zerofill, format1 - len);
+
+		  grub_size_t i;
+		  for (i = 0; i < len; i++)
+		    write_char (*p++);
+
+		  if (rightfill && len < format1)
+		    write_fill (zerofill, format1 - len);
 		}
 	      else
 		write_str ("(null)");
-	      
+
 	      break;
 
 	    default:
@@ -897,109 +875,107 @@ grub_vsprintf (char *str, const char *fmt, va_list args)
   if (str)
     *str = '\0';
 
-  if (count && !str)
-    grub_refresh ();
-  
   return count;
 }
 
 int
-grub_sprintf (char *str, const char *fmt, ...)
+grub_vsnprintf (char *str, grub_size_t n, const char *fmt, va_list ap)
+{
+  grub_size_t ret;
+
+  if (!n)
+    return 0;
+
+  n--;
+
+  ret = grub_vsnprintf_real (str, n, fmt, ap);
+
+  return ret < n ? ret : n;
+}
+
+int
+grub_snprintf (char *str, grub_size_t n, const char *fmt, ...)
 {
   va_list ap;
   int ret;
-  
+
   va_start (ap, fmt);
-  ret = grub_vsprintf (str, fmt, ap);
+  ret = grub_vsnprintf (str, n, fmt, ap);
   va_end (ap);
 
   return ret;
 }
 
-/* Convert UTF-16 to UTF-8.  */
-grub_uint8_t *
-grub_utf16_to_utf8 (grub_uint8_t *dest, grub_uint16_t *src,
-		    grub_size_t size)
+#define PREALLOC_SIZE 255
+
+char *
+grub_xvasprintf (const char *fmt, va_list ap)
 {
-  grub_uint32_t code_high = 0;
+  grub_size_t s, as = PREALLOC_SIZE;
+  char *ret;
 
-  while (size--)
+  while (1)
     {
-      grub_uint32_t code = *src++;
+      ret = grub_malloc (as + 1);
+      if (!ret)
+	return NULL;
 
-      if (code_high)
-	{
-	  if (code >= 0xDC00 && code <= 0xDFFF)
-	    {
-	      /* Surrogate pair.  */
-	      code = ((code_high - 0xD800) << 12) + (code - 0xDC00) + 0x10000;
-	      
-	      *dest++ = (code >> 18) | 0xF0;
-	      *dest++ = ((code >> 12) & 0x3F) | 0x80;
-	      *dest++ = ((code >> 6) & 0x3F) | 0x80;
-	      *dest++ = (code & 0x3F) | 0x80;
-	    }
-	  else
-	    {
-	      /* Error...  */
-	      *dest++ = '?';
-	    }
+      s = grub_vsnprintf_real (ret, as, fmt, ap);
+      if (s <= as)
+	return ret;
 
-	  code_high = 0;
-	}
-      else
-	{
-	  if (code <= 0x007F)
-	    *dest++ = code;
-	  else if (code <= 0x07FF)
-	    {
-	      *dest++ = (code >> 6) | 0xC0;
-	      *dest++ = (code & 0x3F) | 0x80;
-	    }
-	  else if (code >= 0xD800 && code <= 0xDBFF)
-	    {
-	      code_high = code;
-	      continue;
-	    }
-	  else if (code >= 0xDC00 && code <= 0xDFFF)
-	    {
-	      /* Error... */
-	      *dest++ = '?';
-	    }
-	  else
-	    {
-	      *dest++ = (code >> 12) | 0xE0;
-	      *dest++ = ((code >> 6) & 0x3F) | 0x80;
-	      *dest++ = (code & 0x3F) | 0x80;
-	    }
-	}
+      grub_free (ret);
+      as = s;
     }
-
-  return dest;
 }
 
-/* Convert an UTF-8 string to an UCS-4 string. Return the number of
-   characters converted. DEST must be able to hold at least SIZE
-   characters (when the input is unknown). If an invalid sequence is found,
-   return -1.  */
-grub_ssize_t
-grub_utf8_to_ucs4 (grub_uint32_t *dest, const grub_uint8_t *src,
-		   grub_size_t size)
+char *
+grub_xasprintf (const char *fmt, ...)
+{
+  va_list ap;
+  char *ret;
+
+  va_start (ap, fmt);
+  ret = grub_xvasprintf (fmt, ap);
+  va_end (ap);
+
+  return ret;
+}
+
+/* Convert a (possibly null-terminated) UTF-8 string of at most SRCSIZE
+   bytes (if SRCSIZE is -1, it is ignored) in length to a UCS-4 string.
+   Return the number of characters converted. DEST must be able to hold
+   at least DESTSIZE characters.
+   If SRCEND is not NULL, then *SRCEND is set to the next byte after the
+   last byte used in SRC.  */
+grub_size_t
+grub_utf8_to_ucs4 (grub_uint32_t *dest, grub_size_t destsize,
+		   const grub_uint8_t *src, grub_size_t srcsize,
+		   const grub_uint8_t **srcend)
 {
   grub_uint32_t *p = dest;
   int count = 0;
   grub_uint32_t code = 0;
-  
-  while (size--)
+
+  if (srcend)
+    *srcend = src;
+
+  while (srcsize && destsize)
     {
       grub_uint32_t c = *src++;
-      
+      if (srcsize != (grub_size_t)-1)
+	srcsize--;
       if (count)
 	{
 	  if ((c & 0xc0) != 0x80)
 	    {
 	      /* invalid */
-	      return -1;
+	      code = '?';
+	      /* Character c may be valid, don't eat it.  */
+	      src--;
+	      if (srcsize != (grub_size_t)-1)
+		srcsize++;
+	      count = 0;
 	    }
 	  else
 	    {
@@ -1010,6 +986,9 @@ grub_utf8_to_ucs4 (grub_uint32_t *dest, const grub_uint8_t *src,
 	}
       else
 	{
+	  if (c == 0)
+	    break;
+
 	  if ((c & 0x80) == 0x00)
 	    code = c;
 	  else if ((c & 0xe0) == 0xc0)
@@ -1038,39 +1017,52 @@ grub_utf8_to_ucs4 (grub_uint32_t *dest, const grub_uint8_t *src,
 	      code = c & 0x01;
 	    }
 	  else
-	    /* invalid */
-	    return -1;
+	    {
+	      /* invalid */
+	      code = '?';
+	      count = 0;
+	    }
 	}
 
       if (count == 0)
-	*p++ = code;
+	{
+	  *p++ = code;
+	  destsize--;
+	}
     }
 
+  if (srcend)
+    *srcend = src;
   return p - dest;
 }
 
-void
-grub_millisleep_generic (grub_uint32_t ms)
-{
-  grub_uint32_t end_at;
-
-  end_at = grub_get_rtc () + grub_div_roundup (ms * GRUB_TICKS_PER_SECOND, 1000);
-
-  while (grub_get_rtc () < end_at)
-    grub_cpu_idle ();
-}
-
 /* Abort GRUB. This function does not return.  */
 void
 grub_abort (void)
 {
-  if (grub_term_get_current ())
+  grub_printf ("\nAborted.");
+  
+#ifndef GRUB_UTIL
+  if (grub_term_inputs)
+#endif
     {
-      grub_printf ("\nAborted. Press any key to exit.");
+      grub_printf (" Press any key to exit.");
       grub_getkey ();
     }
 
   grub_exit ();
 }
+
+#ifndef APPLE_CC
 /* GCC emits references to abort().  */
 void abort (void) __attribute__ ((alias ("grub_abort")));
+#endif
+
+#ifdef NEED_ENABLE_EXECUTE_STACK
+/* Some gcc versions generate a call to this function
+   in trampolines for nested functions.  */
+void __enable_execute_stack (void *addr __attribute__ ((unused)))
+{
+}
+#endif
+
