@@ -1,7 +1,7 @@
 /* chainloader.c - boot another boot loader */
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 2002,2004,2006,2007  Free Software Foundation, Inc.
+ *  Copyright (C) 2002,2004,2006,2007,2008  Free Software Foundation, Inc.
  *
  *  GRUB is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -25,14 +25,15 @@
 #include <grub/device.h>
 #include <grub/disk.h>
 #include <grub/misc.h>
+#include <grub/charset.h>
 #include <grub/mm.h>
 #include <grub/types.h>
-#include <grub/rescue.h>
 #include <grub/dl.h>
 #include <grub/efi/api.h>
 #include <grub/efi/efi.h>
 #include <grub/efi/disk.h>
-#include <grub/efi/chainloader.h>
+#include <grub/command.h>
+#include <grub/i18n.h>
 
 static grub_dl_t my_mod;
 
@@ -40,17 +41,21 @@ static grub_efi_physical_address_t address;
 static grub_efi_uintn_t pages;
 static grub_efi_device_path_t *file_path;
 static grub_efi_handle_t image_handle;
+static grub_efi_char16_t *cmdline;
 
 static grub_err_t
 grub_chainloader_unload (void)
 {
   grub_efi_boot_services_t *b;
-  
+
   b = grub_efi_system_table->boot_services;
-  b->unload_image (image_handle);
-  b->free_pages (address, pages);
+  efi_call_1 (b->unload_image, image_handle);
+  efi_call_2 (b->free_pages, address, pages);
+
   grub_free (file_path);
-  
+  grub_free (cmdline);
+  cmdline = 0;
+
   grub_dl_unref (my_mod);
   return GRUB_ERR_NONE;
 }
@@ -62,21 +67,21 @@ grub_chainloader_boot (void)
   grub_efi_status_t status;
   grub_efi_uintn_t exit_data_size;
   grub_efi_char16_t *exit_data;
-  
+
   b = grub_efi_system_table->boot_services;
-  status = b->start_image (image_handle, &exit_data_size, &exit_data);
+  status = efi_call_3 (b->start_image, image_handle, &exit_data_size, &exit_data);
   if (status != GRUB_EFI_SUCCESS)
     {
       if (exit_data)
 	{
 	  char *buf;
-	  
+
 	  buf = grub_malloc (exit_data_size * 4 + 1);
 	  if (buf)
 	    {
 	      *grub_utf16_to_utf8 ((grub_uint8_t *) buf,
 				   exit_data, exit_data_size) = 0;
-	      
+
 	      grub_error (GRUB_ERR_BAD_OS, buf);
 	      grub_free (buf);
 	    }
@@ -86,10 +91,10 @@ grub_chainloader_boot (void)
     }
 
   if (exit_data)
-    b->free_pool (exit_data);
+    efi_call_1 (b->free_pool, exit_data);
 
   grub_chainloader_unload ();
-  
+
   return grub_errno;
 }
 
@@ -99,7 +104,7 @@ copy_file_path (grub_efi_file_path_device_path_t *fp,
 {
   grub_efi_char16_t *p;
   grub_efi_uint16_t size;
-  
+
   fp->header.type = GRUB_EFI_MEDIA_DEVICE_PATH_TYPE;
   fp->header.subtype = GRUB_EFI_FILE_PATH_DEVICE_PATH_SUBTYPE;
   size = len * sizeof (grub_efi_char16_t) + sizeof (*fp);
@@ -132,7 +137,7 @@ make_file_path (grub_efi_device_path_t *dp, const char *filename)
       grub_error (GRUB_ERR_BAD_FILENAME, "invalid EFI file path");
       return 0;
     }
-  
+
   size = 0;
   d = dp;
   while (1)
@@ -142,7 +147,7 @@ make_file_path (grub_efi_device_path_t *dp, const char *filename)
 	break;
       d = GRUB_EFI_NEXT_DEVICE_PATH (d);
     }
-  
+
   file_path = grub_malloc (size
 			   + ((grub_strlen (dir_start) + 1)
 			      * sizeof (grub_efi_char16_t))
@@ -151,7 +156,7 @@ make_file_path (grub_efi_device_path_t *dp, const char *filename)
     return 0;
 
   grub_memcpy (file_path, dp, size);
-  
+
   /* Fill the file path for the directory.  */
   d = (grub_efi_device_path_t *) ((char *) file_path
 				  + ((char *) d - (char *) dp));
@@ -174,8 +179,9 @@ make_file_path (grub_efi_device_path_t *dp, const char *filename)
   return file_path;
 }
 
-void
-grub_chainloader_cmd (const char *filename)
+static grub_err_t
+grub_cmd_chainloader (grub_command_t cmd __attribute__ ((unused)),
+		      int argc, char *argv[])
 {
   grub_file_t file = 0;
   grub_ssize_t size;
@@ -185,14 +191,19 @@ grub_chainloader_cmd (const char *filename)
   grub_device_t dev = 0;
   grub_efi_device_path_t *dp = 0;
   grub_efi_loaded_image_t *loaded_image;
-  
+  char *filename;
+
+  if (argc == 0)
+    return grub_error (GRUB_ERR_BAD_ARGUMENT, "no file specified");
+  filename = argv[0];
+
   grub_dl_ref (my_mod);
 
   /* Initialize some global variables.  */
   address = 0;
   image_handle = 0;
   file_path = 0;
-  
+
   b = grub_efi_system_table->boot_services;
 
   file = grub_file_open (filename);
@@ -210,7 +221,7 @@ grub_chainloader_cmd (const char *filename)
       if (dev_handle)
 	dp = grub_efi_get_device_path (dev_handle);
     }
-  
+
   if (! dev->disk || ! dev_handle || ! dp)
     {
       grub_error (GRUB_ERR_BAD_DEVICE, "not a valid root device");
@@ -223,11 +234,11 @@ grub_chainloader_cmd (const char *filename)
 
   grub_printf ("file path: ");
   grub_efi_print_device_path (file_path);
-  
+
   size = grub_file_size (file);
   pages = (((grub_efi_uintn_t) size + ((1 << 12) - 1)) >> 12);
-  
-  status = b->allocate_pages (GRUB_EFI_ALLOCATE_ANY_PAGES,
+
+  status = efi_call_4 (b->allocate_pages, GRUB_EFI_ALLOCATE_ANY_PAGES,
 			      GRUB_EFI_LOADER_CODE,
 			      pages, &address);
   if (status != GRUB_EFI_SUCCESS)
@@ -244,7 +255,7 @@ grub_chainloader_cmd (const char *filename)
       goto fail;
     }
 
-  status = b->load_image (0, grub_efi_image_handle, file_path,
+  status = efi_call_6 (b->load_image, 0, grub_efi_image_handle, file_path,
 			  (void *) ((grub_addr_t) address), size,
 			  &image_handle);
   if (status != GRUB_EFI_SUCCESS)
@@ -253,7 +264,7 @@ grub_chainloader_cmd (const char *filename)
 	grub_error (GRUB_ERR_OUT_OF_MEMORY, "out of resources");
       else
 	grub_error (GRUB_ERR_BAD_OS, "cannot load image");
-      
+
       goto fail;
     }
 
@@ -267,48 +278,70 @@ grub_chainloader_cmd (const char *filename)
       goto fail;
     }
   loaded_image->device_handle = dev_handle;
-  
+
   grub_file_close (file);
+
+  if (argc > 1)
+    {
+      int i, len;
+      grub_efi_char16_t *p16;
+
+      for (i = 1, len = 0; i < argc; i++)
+        len += grub_strlen (argv[i]) + 1;
+
+      len *= sizeof (grub_efi_char16_t);
+      cmdline = p16 = grub_malloc (len);
+      if (! cmdline)
+        goto fail;
+
+      for (i = 1; i < argc; i++)
+        {
+          char *p8;
+
+          p8 = argv[i];
+          while (*p8)
+            *(p16++) = *(p8++);
+
+          *(p16++) = ' ';
+        }
+      *(--p16) = 0;
+
+      loaded_image->load_options = cmdline;
+      loaded_image->load_options_size = len;
+    }
+
   grub_loader_set (grub_chainloader_boot, grub_chainloader_unload, 0);
-  return;
-  
+  return 0;
+
  fail:
 
   if (dev)
     grub_device_close (dev);
-  
+
   if (file)
     grub_file_close (file);
 
   if (file_path)
     grub_free (file_path);
-  
+
   if (address)
-    b->free_pages (address, pages);
-  
+    efi_call_2 (b->free_pages, address, pages);
+
   grub_dl_unref (my_mod);
+
+  return grub_errno;
 }
 
-static void
-grub_rescue_cmd_chainloader (int argc, char *argv[])
-{
-  if (argc == 0)
-    grub_error (GRUB_ERR_BAD_ARGUMENT, "no file specified");
-  else
-    grub_chainloader_cmd (argv[0]);
-}
-
-static const char loader_name[] = "chainloader";
+static grub_command_t cmd;
 
 GRUB_MOD_INIT(chainloader)
 {
-  grub_rescue_register_command (loader_name,
-				grub_rescue_cmd_chainloader,
-				"load another boot loader");
+  cmd = grub_register_command ("chainloader", grub_cmd_chainloader,
+			       0, N_("Load another boot loader."));
   my_mod = mod;
 }
 
 GRUB_MOD_FINI(chainloader)
 {
-  grub_rescue_unregister_command (loader_name);
+  grub_unregister_command (cmd);
 }
