@@ -102,7 +102,7 @@ amiga_partition_map_iterate (grub_disk_t disk,
 
   if (next == -1)
     return grub_error (GRUB_ERR_BAD_PART_TABLE,
-		       "Amiga partition map not found.");
+		       "Amiga partition map not found");
 
   /* The end of the partition list is marked using "-1".  */
   while (next != -1)
@@ -184,21 +184,14 @@ amiga_partition_map_probe (grub_disk_t disk, const char *str)
 static char *
 amiga_partition_map_get_name (const grub_partition_t p)
 {
-  char *name;
-
-  name = grub_malloc (13);
-  if (! name)
-    return 0;
-
-  grub_sprintf (name, "%d", p->index + 1);
-  return name;
+  return grub_xasprintf ("%d", p->index + 1);
 }
 
 
 /* Partition map type.  */
 static struct grub_partition_map grub_amiga_partition_map =
   {
-    .name = "amiga_partition_map",
+    .name = "part_amiga",
     .iterate = amiga_partition_map_iterate,
     .probe = amiga_partition_map_probe,
     .get_name = amiga_partition_map_get_name
