@@ -1,7 +1,7 @@
 /* apple.c - Read macintosh partition tables.  */
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 2002,2004,2005,2006,2007,2008  Free Software Foundation, Inc.
+ *  Copyright (C) 2002,2004,2005,2006,2007,2008,2009  Free Software Foundation, Inc.
  *
  *  GRUB is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -30,6 +30,7 @@ struct grub_apple_header
   /* The magic number to identify the partition map, it should have
      the value `0x4552'.  */
   grub_uint16_t magic;
+  grub_uint16_t blocksize;
 };
 
 struct grub_apple_part
@@ -105,8 +106,8 @@ apple_partition_map_iterate (grub_disk_t disk,
   struct grub_apple_header aheader;
   struct grub_apple_part apart;
   struct grub_disk raw;
-  int partno = 0;
-  unsigned pos = GRUB_DISK_SECTOR_SIZE;
+  int partno = 0, partnum = 0;
+  unsigned pos;
 
   /* Enforce raw disk access.  */
   raw = *disk;
@@ -126,7 +127,9 @@ apple_partition_map_iterate (grub_disk_t disk,
       goto fail;
     }
 
-  for (;;)
+  pos = grub_be_to_cpu16 (aheader.blocksize);
+
+  do
     {
       if (grub_disk_read (&raw, pos / GRUB_DISK_SECTOR_SIZE,
 			  pos % GRUB_DISK_SECTOR_SIZE,
@@ -142,8 +145,15 @@ apple_partition_map_iterate (grub_disk_t disk,
 	  break;
 	}
 
-      part.start = grub_be_to_cpu32 (apart.first_phys_block);
-      part.len = grub_be_to_cpu32 (apart.blockcnt);
+      if (partnum == 0)
+	partnum = grub_be_to_cpu32 (apart.partmap_size);
+
+      part.start = ((grub_disk_addr_t) grub_be_to_cpu32 (apart.first_phys_block)
+		    * grub_be_to_cpu16 (aheader.blocksize))
+	/ GRUB_DISK_SECTOR_SIZE;
+      part.len = ((grub_disk_addr_t) grub_be_to_cpu32 (apart.blockcnt)
+		  * grub_be_to_cpu16 (aheader.blocksize))
+	/ GRUB_DISK_SECTOR_SIZE;
       part.offset = pos;
       part.index = partno;
 
@@ -156,20 +166,17 @@ apple_partition_map_iterate (grub_disk_t disk,
       if (hook (disk, &part))
 	return grub_errno;
 
-      if (grub_be_to_cpu32 (apart.first_phys_block)
-	  == GRUB_DISK_SECTOR_SIZE * 2)
-	return 0;
-
-      pos += sizeof (struct grub_apple_part);
+      pos += grub_be_to_cpu16 (aheader.blocksize);
       partno++;
     }
+  while (partno < partnum);
 
-  if (pos != GRUB_DISK_SECTOR_SIZE)
+  if (partno != 0)
     return 0;
 
  fail:
   return grub_error (GRUB_ERR_BAD_PART_TABLE,
-		     "Apple partition map not found.");
+		     "Apple partition map not found");
 }
 
 
@@ -220,21 +227,14 @@ apple_partition_map_probe (grub_disk_t disk, const char *str)
 static char *
 apple_partition_map_get_name (const grub_partition_t p)
 {
-  char *name;
-
-  name = grub_malloc (13);
-  if (! name)
-    return 0;
-
-  grub_sprintf (name, "%d", p->index + 1);
-  return name;
+  return grub_xasprintf ("%d", p->index + 1);
 }
 
 
 /* Partition map type.  */
 static struct grub_partition_map grub_apple_partition_map =
   {
-    .name = "apple_partition_map",
+    .name = "part_apple",
     .iterate = apple_partition_map_iterate,
     .probe = apple_partition_map_probe,
     .get_name = apple_partition_map_get_name
