@@ -1,7 +1,7 @@
 /* gpt.c - Read GUID Partition Tables (GPT).  */
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 2002,2005,2006,2007  Free Software Foundation, Inc.
+ *  Copyright (C) 2002,2005,2006,2007,2008  Free Software Foundation, Inc.
  *
  *  GRUB is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -22,49 +22,18 @@
 #include <grub/mm.h>
 #include <grub/partition.h>
 #include <grub/dl.h>
-#include <grub/pc_partition.h>
-
-struct grub_gpt_header
-{
-  grub_uint8_t magic[8];
-  grub_uint32_t version;
-  grub_uint32_t headersize;
-  grub_uint32_t crc32;
-  grub_uint32_t unused1;
-  grub_uint64_t primary;
-  grub_uint64_t backup;
-  grub_uint64_t start;
-  grub_uint64_t end;
-  grub_uint8_t guid[16];
-  grub_uint64_t partitions;
-  grub_uint32_t maxpart;
-  grub_uint32_t partentry_size;
-  grub_uint32_t partentry_crc32;
-} __attribute__ ((packed));
-
-struct grub_gpt_partentry
-{
-  grub_uint8_t type[16];
-  grub_uint8_t guid[16];
-  grub_uint64_t start;
-  grub_uint64_t end;
-  grub_uint8_t attrib;
-  char name[72];
-} __attribute__ ((packed));
+#include <grub/msdos_partition.h>
+#include <grub/gpt_partition.h>
 
 static grub_uint8_t grub_gpt_magic[8] =
   {
-    45, 46, 49, 20, 50, 41, 52, 54
+    0x45, 0x46, 0x49, 0x20, 0x50, 0x41, 0x52, 0x54
   };
 
-static grub_uint8_t grub_gpt_partition_type_empty[16] = { 0 };
+static const grub_gpt_part_type_t grub_gpt_partition_type_empty = GRUB_GPT_PARTITION_TYPE_EMPTY;
 
 static struct grub_partition_map grub_gpt_partition_map;
 
-#ifndef GRUB_UTIL
-static grub_dl_t my_mod;
-#endif
-
 
 
 static grub_err_t
@@ -76,9 +45,8 @@ gpt_partition_map_iterate (grub_disk_t disk,
   struct grub_gpt_header gpt;
   struct grub_gpt_partentry entry;
   struct grub_disk raw;
-  struct grub_pc_partition_mbr mbr;
+  struct grub_msdos_partition_mbr mbr;
   grub_uint64_t entries;
-  int partno = 0;
   unsigned int i;
   int last_offset = 0;
 
@@ -87,7 +55,7 @@ gpt_partition_map_iterate (grub_disk_t disk,
   raw.partition = 0;
 
   /* Read the protective MBR.  */
-  if (grub_disk_read (&raw, 0, 0, sizeof (mbr), (char *) &mbr))
+  if (grub_disk_read (&raw, 0, 0, sizeof (mbr), &mbr))
     return grub_errno;
 
   /* Check if it is valid.  */
@@ -99,10 +67,10 @@ gpt_partition_map_iterate (grub_disk_t disk,
     return grub_error (GRUB_ERR_BAD_PART_TABLE, "no GPT partition map found");
 
   /* Read the GPT header.  */
-  if (grub_disk_read (&raw, 1, 0, sizeof (gpt), (char *) &gpt))
+  if (grub_disk_read (&raw, 1, 0, sizeof (gpt), &gpt))
     return grub_errno;
 
-  if (! grub_memcmp (gpt.magic, grub_gpt_magic, sizeof (grub_gpt_magic)))
+  if (grub_memcmp (gpt.magic, grub_gpt_magic, sizeof (grub_gpt_magic)))
     return grub_error (GRUB_ERR_BAD_PART_TABLE, "no valid GPT header");
 
   grub_dprintf ("gpt", "Read a valid GPT header\n");
@@ -111,10 +79,10 @@ gpt_partition_map_iterate (grub_disk_t disk,
   for (i = 0; i < grub_le_to_cpu32 (gpt.maxpart); i++)
     {
       if (grub_disk_read (&raw, entries, last_offset,
-			  sizeof (entry), (char *) &entry))
+			  sizeof (entry), &entry))
 	return grub_errno;
 
-      if (grub_memcmp (grub_gpt_partition_type_empty, entry.type,
+      if (grub_memcmp (&grub_gpt_partition_type_empty, &entry.type,
 		       sizeof (grub_gpt_partition_type_empty)))
 	{
 	  /* Calculate the first block and the size of the partition.  */
@@ -122,17 +90,18 @@ gpt_partition_map_iterate (grub_disk_t disk,
 	  part.len = (grub_le_to_cpu64 (entry.end)
 		      - grub_le_to_cpu64 (entry.start) + 1);
 	  part.offset = entries;
-	  part.index = partno;
+	  part.index = i;
 	  part.partmap = &grub_gpt_partition_map;
+	  part.data = &entry;
 
-	  grub_dprintf ("gpt", "GPT entry %d: start=%lld, length=%lld\n",
-			partno, part.start, part.len);
+	  grub_dprintf ("gpt", "GPT entry %d: start=%lld, length=%lld\n", i,
+			(unsigned long long) part.start,
+			(unsigned long long) part.len);
 
 	  if (hook (disk, &part))
-	    return grub_errno;
+	    return 1;
 	}
 
-      partno++;
       last_offset += grub_le_to_cpu32 (gpt.partentry_size);
       if (last_offset == GRUB_DISK_SECTOR_SIZE)
 	{
@@ -153,7 +122,7 @@ gpt_partition_map_probe (grub_disk_t disk, const char *str)
   char *s = (char *) str;
 
   auto int find_func (grub_disk_t d, const grub_partition_t partition);
-    
+
   int find_func (grub_disk_t d __attribute__ ((unused)),
 		 const grub_partition_t partition)
     {
@@ -162,14 +131,14 @@ gpt_partition_map_probe (grub_disk_t disk, const char *str)
 	  p = (grub_partition_t) grub_malloc (sizeof (*p));
 	  if (! p)
 	    return 1;
-	  
+
 	  grub_memcpy (p, partition, sizeof (*p));
 	  return 1;
 	}
-      
+
       return 0;
     }
-  
+
   /* Get the partition number.  */
   partnum = grub_strtoul (s, 0, 10) - 1;
   if (grub_errno)
@@ -178,7 +147,8 @@ gpt_partition_map_probe (grub_disk_t disk, const char *str)
       return 0;
     }
 
-  if (gpt_partition_map_iterate (disk, find_func))
+  gpt_partition_map_iterate (disk, find_func);
+  if (grub_errno)
     goto fail;
 
   return p;
@@ -192,21 +162,14 @@ gpt_partition_map_probe (grub_disk_t disk, const char *str)
 static char *
 gpt_partition_map_get_name (const grub_partition_t p)
 {
-  char *name;
-
-  name = grub_malloc (13);
-  if (! name)
-    return 0;
-
-  grub_sprintf (name, "%d", p->index + 1);
-  return name;
+  return grub_xasprintf ("%d", p->index + 1);
 }
 
 
 /* Partition map type.  */
 static struct grub_partition_map grub_gpt_partition_map =
   {
-    .name = "gpt_partition_map",
+    .name = "part_gpt",
     .iterate = gpt_partition_map_iterate,
     .probe = gpt_partition_map_probe,
     .get_name = gpt_partition_map_get_name
@@ -215,9 +178,6 @@ static struct grub_partition_map grub_gpt_partition_map =
 GRUB_MOD_INIT(gpt_partition_map)
 {
   grub_partition_map_register (&grub_gpt_partition_map);
-#ifndef GRUB_UTIL
-  my_mod = mod;
-#endif
 }
 
 GRUB_MOD_FINI(gpt_partition_map)
