@@ -1,7 +1,7 @@
 /* grub-fstest.c - debug tool for filesystem driver */
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 2008 Free Software Foundation, Inc.
+ *  Copyright (C) 2008,2009,2010 Free Software Foundation, Inc.
  *
  *  GRUB is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -28,10 +28,10 @@
 #include <grub/env.h>
 #include <grub/term.h>
 #include <grub/mm.h>
-#include <grub/raid.h>
 #include <grub/lib/hexdump.h>
 #include <grub/lib/crc.h>
 #include <grub/command.h>
+#include <grub/i18n.h>
 
 #include <grub_fstest_init.h>
 
@@ -41,6 +41,8 @@
 #include <stdlib.h>
 #include <getopt.h>
 
+#include "progname.h"
+
 void
 grub_putchar (int c)
 {
@@ -69,7 +71,7 @@ execute_command (char *name, int n, char **args)
 
   cmd = grub_command_find (name);
   if (! cmd)
-    grub_util_error ("Can\'t find command %s", name);
+    grub_util_error ("can\'t find command %s", name);
 
   return (cmd->func) (cmd, n, args);
 }
@@ -83,7 +85,7 @@ execute_command (char *name, int n, char **args)
 
 #define BUF_SIZE  32256
 
-static grub_off_t skip, leng;
+static grub_disk_addr_t skip, leng;
 
 static void
 read_file (char *pathname, int (*hook) (grub_off_t ofs, char *buf, int len))
@@ -98,9 +100,9 @@ read_file (char *pathname, int (*hook) (grub_off_t ofs, char *buf, int len))
 
       dev = grub_device_open (0);
       if ((! dev) || (! dev->disk))
-        grub_util_error ("Can\'t open device.");
+        grub_util_error ("can\'t open device");
 
-      grub_util_info ("total sectors : %lld.",
+      grub_util_info ("total sectors : %lld",
                       (unsigned long long) dev->disk->total_sectors);
 
       if (! leng)
@@ -113,7 +115,7 @@ read_file (char *pathname, int (*hook) (grub_off_t ofs, char *buf, int len))
           len = (leng > BUF_SIZE) ? BUF_SIZE : leng;
 
           if (grub_disk_read (dev->disk, 0, skip, len, buf))
-            grub_util_error ("Disk read fails at offset %lld, length %d.",
+            grub_util_error ("disk read fails at offset %lld, length %d",
                              skip, len);
 
           if (hook (skip, buf, len))
@@ -130,15 +132,15 @@ read_file (char *pathname, int (*hook) (grub_off_t ofs, char *buf, int len))
   file = grub_file_open (pathname);
   if (!file)
     {
-      grub_util_error ("cannot open file %s.", pathname);
+      grub_util_error ("cannot open file %s", pathname);
       return;
     }
 
-  grub_util_info ("file size : %lld.", (unsigned long long) file->size);
+  grub_util_info ("file size : %lld", (unsigned long long) file->size);
 
   if (skip > file->size)
     {
-      grub_util_error ("invalid skip value %d.");
+      grub_util_error ("invalid skip value %lld", (unsigned long long) skip);
       return;
     }
 
@@ -156,7 +158,7 @@ read_file (char *pathname, int (*hook) (grub_off_t ofs, char *buf, int len))
       sz = grub_file_read (file, buf, (len > BUF_SIZE) ? BUF_SIZE : len);
       if (sz < 0)
 	{
-	  grub_util_error ("read error at offset %llu.", ofs);
+	  grub_util_error ("read error at offset %llu", ofs);
 	  break;
 	}
 
@@ -182,7 +184,7 @@ cmd_cp (char *src, char *dest)
 
     if ((int) fwrite (buf, 1, len, ff) != len)
       {
-	grub_util_error ("write error.");
+	grub_util_error ("write error");
 	return 1;
       }
 
@@ -192,7 +194,7 @@ cmd_cp (char *src, char *dest)
   ff = fopen (dest, "wb");
   if (ff == NULL)
     {
-      grub_util_error ("open error.");
+      grub_util_error ("open error");
       return;
     }
   read_file (src, cp_hook);
@@ -210,7 +212,7 @@ cmd_cmp (char *src, char *dest)
   {
     if ((int) fread (buf_1, 1, len, ff) != len)
       {
-	grub_util_error ("read error at offset %llu.", ofs);
+	grub_util_error ("read error at offset %llu", ofs);
 	return 1;
       }
 
@@ -221,7 +223,7 @@ cmd_cmp (char *src, char *dest)
 	for (i = 0; i < len; i++, ofs++)
 	  if (buf_1[i] != buf[i])
 	    {
-	      grub_util_error ("compare fail at offset %llu.", ofs);
+	      grub_util_error ("compare fail at offset %llu", ofs);
 	      return 1;
 	    }
       }
@@ -231,12 +233,12 @@ cmd_cmp (char *src, char *dest)
   ff = fopen (dest, "rb");
   if (ff == NULL)
     {
-      grub_util_error ("open error.");
+      grub_util_error ("open error");
       return;
     }
 
   if ((skip) && (fseeko (ff, skip, SEEK_SET)))
-    grub_util_error ("seek error.");
+    grub_util_error ("seek error");
 
   read_file (src, cmp_hook);
   fclose (ff);
@@ -276,24 +278,40 @@ cmd_crc (char *pathname)
 static void
 fstest (char **images, int num_disks, int cmd, int n, char **args)
 {
-  char host_file[128];
-  char loop_name[8];
-  char *argv[3] = { "-p", loop_name, host_file};
+  char *host_file;
+  char *loop_name;
+  char *argv[3];
   int i;
 
+  argv[0] = "-p";
+
   for (i = 0; i < num_disks; i++)
     {
-      if (grub_strlen (images[i]) + 7 > sizeof (host_file))
-        grub_util_error ("Pathname %s too long.", images[i]);
+      loop_name = grub_xasprintf ("loop%d", i);
+      if (!loop_name)
+	grub_util_error (grub_errmsg);
 
-      grub_sprintf (loop_name, "loop%d", i);
-      grub_sprintf (host_file, "(host)%s", images[i]);
+      host_file = grub_xasprintf ("(host)%s", images[i]);
+      if (!host_file)
+	grub_util_error (grub_errmsg);
+
+      argv[1] = loop_name;
+      argv[2] = host_file;
 
       if (execute_command ("loopback", 3, argv))
-        grub_util_error ("loopback command fails.");
+        grub_util_error ("loopback command fails");
+
+      grub_free (loop_name);
+      grub_free (host_file);
     }
 
-  grub_raid_rescan ();
+  grub_lvm_fini ();
+  grub_mdraid_fini ();
+  grub_raid_fini ();
+  grub_raid_init ();
+  grub_mdraid_init ();
+  grub_lvm_init ();
+
   switch (cmd)
     {
     case CMD_LS:
@@ -320,8 +338,15 @@ fstest (char **images, int num_disks, int cmd, int n, char **args)
 
   for (i = 0; i < num_disks; i++)
     {
-      grub_sprintf (loop_name, "loop%d", i);
+      loop_name = grub_xasprintf ("loop%d", i);
+      if (!loop_name)
+	grub_util_error (grub_errmsg);
+
+      argv[1] = loop_name;
+
       execute_command ("loopback", 2, argv);
+
+      grub_free (loop_name);
     }
 }
 
@@ -341,10 +366,10 @@ static void
 usage (int status)
 {
   if (status)
-    fprintf (stderr, "Try ``grub-fstest --help'' for more information.\n");
+    fprintf (stderr, "Try `%s --help' for more information.\n", program_name);
   else
     printf ("\
-Usage: grub-fstest [OPTION]... IMAGE_PATH COMMANDS\n\
+Usage: %s [OPTION]... IMAGE_PATH COMMANDS\n\
 \n\
 Debug tool for filesystem driver.\n\
 \nCommands:\n\
@@ -364,7 +389,7 @@ Debug tool for filesystem driver.\n\
   -V, --version             print version information and exit\n\
   -v, --verbose             print verbose messages\n\
 \n\
-Report bugs to <%s>.\n", PACKAGE_BUGREPORT);
+Report bugs to <%s>.\n", program_name, PACKAGE_BUGREPORT);
 
   exit (status);
 }
@@ -372,10 +397,12 @@ Report bugs to <%s>.\n", PACKAGE_BUGREPORT);
 int
 main (int argc, char *argv[])
 {
-  char *debug_str = 0, *root = 0, *default_root, *alloc_root;
+  char *debug_str = NULL, *root = NULL, *default_root, *alloc_root;
   int i, cmd, num_opts, image_index, num_disks = 1;
 
-  progname = "grub-fstest";
+  set_program_name (argv[0]);
+
+  grub_util_init_nls ();
 
   /* Find the first non option entry.  */
   for (num_opts = 1; num_opts < argc; num_opts++)
@@ -437,7 +464,7 @@ main (int argc, char *argv[])
 	    break;
 
 	  case 'V':
-	    printf ("%s (%s) %s\n", progname, PACKAGE_NAME, PACKAGE_VERSION);
+	    printf ("%s (%s) %s\n", program_name, PACKAGE_NAME, PACKAGE_VERSION);
 	    return 0;
 
 	  case 'v':
