@@ -18,8 +18,6 @@
 
 #include <config.h>
 #include <grub/types.h>
-#include <grub/misc.h>
-#include <grub/emu/misc.h>
 #include <grub/util/misc.h>
 #include <grub/misc.h>
 #include <grub/i18n.h>
@@ -1170,8 +1168,7 @@ main (int argc, char *argv[])
       font_info.style = ft_face->style_flags;
       font_info.size = size;
 
-      if (FT_Set_Pixel_Sizes (ft_face, size, size))
-	grub_util_error ("can't set %dx%d font size", size, size);
+      FT_Set_Pixel_Sizes (ft_face, size, size);
       add_font (&font_info, ft_face, file_format != PF2);
       FT_Done_Face (ft_face);
     }
