@@ -1,7 +1,7 @@
 /* grub-setup.c - make GRUB usable */
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 1999,2000,2001,2002,2003,2004,2005,2006,2007,2008  Free Software Foundation, Inc.
+ *  Copyright (C) 1999,2000,2001,2002,2003,2004,2005,2006,2007,2008,2009,2010  Free Software Foundation, Inc.
  *
  *  GRUB is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -25,15 +25,17 @@
 #include <grub/file.h>
 #include <grub/fs.h>
 #include <grub/partition.h>
-#include <grub/pc_partition.h>
+#include <grub/msdos_partition.h>
 #include <grub/gpt_partition.h>
 #include <grub/env.h>
 #include <grub/util/hostdisk.h>
 #include <grub/machine/boot.h>
 #include <grub/machine/kernel.h>
 #include <grub/term.h>
+#include <grub/i18n.h>
 #include <grub/util/raid.h>
 #include <grub/util/lvm.h>
+#include <grub/util/getroot.h>
 
 static const grub_gpt_part_type_t grub_gpt_partition_type_bios_boot = GRUB_GPT_PARTITION_TYPE_BIOS_BOOT;
 
@@ -46,7 +48,8 @@ static const grub_gpt_part_type_t grub_gpt_partition_type_bios_boot = GRUB_GPT_P
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <dirent.h>
-#include <grub/util/getroot.h>
+#include <assert.h>
+#include "progname.h"
 
 #define _GNU_SOURCE	1
 #include <getopt.h>
@@ -54,14 +57,6 @@ static const grub_gpt_part_type_t grub_gpt_partition_type_bios_boot = GRUB_GPT_P
 #define DEFAULT_BOOT_FILE	"boot.img"
 #define DEFAULT_CORE_FILE	"core.img"
 
-/* This is the blocklist used in the diskboot image.  */
-struct boot_blocklist
-{
-  grub_uint64_t start;
-  grub_uint16_t len;
-  grub_uint16_t segment;
-} __attribute__ ((packed));
-
 void
 grub_putchar (int c)
 {
@@ -86,9 +81,9 @@ grub_refresh (void)
 static void
 setup (const char *dir,
        const char *boot_file, const char *core_file,
-       const char *root, const char *dest, int must_embed, int force)
+       const char *root, const char *dest, int must_embed, int force, int fs_probe)
 {
-  char *boot_path, *core_path, *core_path_dev;
+  char *boot_path, *core_path, *core_path_dev, *core_path_dev_full;
   char *boot_img, *core_img;
   size_t boot_size, core_size;
   grub_uint16_t core_sectors;
@@ -97,7 +92,7 @@ setup (const char *dir,
   grub_uint8_t *boot_drive;
   grub_disk_addr_t *kernel_sector;
   grub_uint16_t *boot_drive_check;
-  struct boot_blocklist *first_block, *block;
+  struct grub_boot_blocklist *first_block, *block;
   grub_int32_t *install_dos_part, *install_bsd_part;
   grub_int32_t dos_part, bsd_part;
   char *tmp_img;
@@ -121,18 +116,18 @@ setup (const char *dir,
   int NESTED_FUNC_ATTR find_usable_region_msdos (grub_disk_t disk __attribute__ ((unused)),
 						 const grub_partition_t p)
     {
-      struct grub_pc_partition *pcdata = p->data;
+      struct grub_msdos_partition *pcdata = p->data;
 
       /* There's always an embed region, and it starts right after the MBR.  */
       embed_region.start = 1;
 
       /* For its end offset, include as many dummy partitions as we can.  */
-      if (! grub_pc_partition_is_empty (pcdata->dos_type)
-	  && ! grub_pc_partition_is_bsd (pcdata->dos_type)
+      if (! grub_msdos_partition_is_empty (pcdata->dos_type)
+	  && ! grub_msdos_partition_is_bsd (pcdata->dos_type)
 	  && embed_region.end > p->start)
 	embed_region.end = p->start;
 
-      return 1;
+      return 0;
     }
 
   auto int NESTED_FUNC_ATTR find_usable_region_gpt (grub_disk_t disk,
@@ -161,7 +156,7 @@ setup (const char *dir,
 		      sector, offset, length);
 
       if (offset != 0 || length != GRUB_DISK_SECTOR_SIZE)
-	grub_util_error ("The first sector of the core file is not sector-aligned");
+	grub_util_error (_("the first sector of the core file is not sector-aligned"));
 
       first_sector = sector;
     }
@@ -169,13 +164,13 @@ setup (const char *dir,
   void NESTED_FUNC_ATTR save_blocklists (grub_disk_addr_t sector, unsigned offset,
 			unsigned length)
     {
-      struct boot_blocklist *prev = block + 1;
+      struct grub_boot_blocklist *prev = block + 1;
 
       grub_util_info ("saving <%llu,%u,%u> with the segment 0x%x",
 		      sector, offset, length, (unsigned) current_segment);
 
       if (offset != 0 || last_length != GRUB_DISK_SECTOR_SIZE)
-	grub_util_error ("Non-sector-aligned data is found in the core file");
+	grub_util_error (_("non-sector-aligned data is found in the core file"));
 
       if (block != first_block
 	  && (grub_le_to_cpu64 (prev->start)
@@ -189,7 +184,7 @@ setup (const char *dir,
 
 	  block--;
 	  if (block->len)
-	    grub_util_error ("The sectors of the core file are too fragmented");
+	    grub_util_error (_("the sectors of the core file are too fragmented"));
 	}
 
       last_length = length;
@@ -200,7 +195,7 @@ setup (const char *dir,
   boot_path = grub_util_get_path (dir, boot_file);
   boot_size = grub_util_get_image_size (boot_path);
   if (boot_size != GRUB_DISK_SECTOR_SIZE)
-    grub_util_error ("The size of `%s' is not %d",
+    grub_util_error (_("the size of `%s' is not %u"),
 		     boot_path, GRUB_DISK_SECTOR_SIZE);
   boot_img = grub_util_read_image (boot_path);
   free (boot_path);
@@ -217,16 +212,16 @@ setup (const char *dir,
   core_sectors = ((core_size + GRUB_DISK_SECTOR_SIZE - 1)
 		  >> GRUB_DISK_SECTOR_BITS);
   if (core_size < GRUB_DISK_SECTOR_SIZE)
-    grub_util_error ("The size of `%s' is too small", core_path);
+    grub_util_error (_("the size of `%s' is too small"), core_path);
   else if (core_size > 0xFFFF * GRUB_DISK_SECTOR_SIZE)
-    grub_util_error ("The size of `%s' is too large", core_path);
+    grub_util_error (_("the size of `%s' is too large"), core_path);
 
   core_img = grub_util_read_image (core_path);
 
   /* Have FIRST_BLOCK to point to the first blocklist.  */
-  first_block = (struct boot_blocklist *) (core_img
-					   + GRUB_DISK_SECTOR_SIZE
-					   - sizeof (*block));
+  first_block = (struct grub_boot_blocklist *) (core_img
+						+ GRUB_DISK_SECTOR_SIZE
+						- sizeof (*block));
 
   install_dos_part = (grub_int32_t *) (core_img + GRUB_DISK_SECTOR_SIZE
 				       + GRUB_KERNEL_MACHINE_INSTALL_DOS_PART);
@@ -251,6 +246,22 @@ setup (const char *dir,
   if (grub_disk_read (dest_dev->disk, 0, 0, GRUB_DISK_SECTOR_SIZE, tmp_img))
     grub_util_error ("%s", grub_errmsg);
 
+  if (dest_dev->disk->partition && fs_probe)
+    {
+      grub_fs_t fs;
+      fs = grub_fs_probe (dest_dev);
+      if (! fs)
+	grub_util_error (_("unable to identify a filesystem in %s; safety check can't be performed"),
+			 dest_dev->disk->name);
+
+      if (! fs->reserved_first_sector)
+	grub_util_error (_("%s appears to contain a %s filesystem which isn't known to "
+			   "reserve space for DOS-style boot.  Installing GRUB there could "
+			   "result in FILESYSTEM DESTRUCTION if valuable data is overwritten "
+			   "by grub-setup (--skip-fs-probe disables this "
+			   "check, use at your own risk)"), dest_dev->disk->name, fs->name);
+    }
+
   /* Copy the possible DOS BPB.  */
   memcpy (boot_img + GRUB_BOOT_MACHINE_BPB_START,
 	  tmp_img + GRUB_BOOT_MACHINE_BPB_START,
@@ -279,21 +290,21 @@ setup (const char *dir,
       if (root_dev->disk->partition)
 	{
 	  if (strcmp (root_dev->disk->partition->partmap->name,
-		      "pc_partition_map") == 0)
+		      "part_msdos") == 0)
 	    {
-	      struct grub_pc_partition *pcdata =
+	      struct grub_msdos_partition *pcdata =
 		root_dev->disk->partition->data;
 	      dos_part = pcdata->dos_part;
 	      bsd_part = pcdata->bsd_part;
 	    }
 	  else if (strcmp (root_dev->disk->partition->partmap->name,
-			   "gpt_partition_map") == 0)
+			   "part_gpt") == 0)
 	    {
 	      dos_part = root_dev->disk->partition->index;
 	      bsd_part = -1;
 	    }
 	  else
-	    grub_util_error ("No PC style partitions found");
+	    grub_util_error (_("no DOS-style partitions found"));
 	}
       else
 	dos_part = bsd_part = -1;
@@ -309,13 +320,13 @@ setup (const char *dir,
 
   if (! dest_dev->disk->has_partitions)
     {
-      grub_util_warn ("Attempting to install GRUB to a partitionless disk.  This is a BAD idea.");
+      grub_util_warn (_("Attempting to install GRUB to a partitionless disk.  This is a BAD idea."));
       goto unable_to_embed;
     }
 
   if (dest_dev->disk->partition)
     {
-      grub_util_warn ("Attempting to install GRUB to a partition instead of the MBR.  This is a BAD idea.");
+      grub_util_warn (_("Attempting to install GRUB to a partition instead of the MBR.  This is a BAD idea."));
       goto unable_to_embed;
     }
 
@@ -334,45 +345,48 @@ setup (const char *dir,
 
   if (! dest_partmap)
     {
-      grub_util_warn ("Attempting to install GRUB to a partitionless disk.  This is a BAD idea.");
+      grub_util_warn (_("Attempting to install GRUB to a partitionless disk.  This is a BAD idea."));
       goto unable_to_embed;
     }
 
-  grub_partition_iterate (dest_dev->disk, (strcmp (dest_partmap, "pc_partition_map") ?
-					   find_usable_region_gpt : find_usable_region_msdos));
+  if (strcmp (dest_partmap, "part_msdos") == 0)
+    grub_partition_iterate (dest_dev->disk, find_usable_region_msdos);
+  else if (strcmp (dest_partmap, "part_gpt") == 0)
+    grub_partition_iterate (dest_dev->disk, find_usable_region_gpt);
+  else
+    grub_util_error (_("No DOS-style partitions found"));
 
   if (embed_region.end == embed_region.start)
     {
-      if (! strcmp (dest_partmap, "pc_partition_map"))
-	grub_util_warn ("This msdos-style partition label has no post-MBR gap; embedding won't be possible!");
+      if (! strcmp (dest_partmap, "part_msdos"))
+	grub_util_warn (_("This msdos-style partition label has no post-MBR gap; embedding won't be possible!"));
       else
-	grub_util_warn ("This GPT partition label has no BIOS Boot Partition; embedding won't be possible!");
+	grub_util_warn (_("This GPT partition label has no BIOS Boot Partition; embedding won't be possible!"));
       goto unable_to_embed;
     }
 
   if ((unsigned long) core_sectors > embed_region.end - embed_region.start)
     {
-      if (core_sectors > 62 * 512)
-	grub_util_warn ("Your core.img is unusually large.  It won't fit in the embedding area.");
-      else if (embed_region.end - embed_region.start < 62 * 512)
-	grub_util_warn ("Your embedding area is unusually small.  core.img won't fit in it.");
-      else
-	grub_util_warn ("Embedding area is too small for core.img.");
+      if (core_sectors > 62)
+	grub_util_warn (_("Your core.img is unusually large.  It won't fit in the embedding area."));
+      else /* embed_region.end - embed_region.start < 62 */
+	grub_util_warn (_("Your embedding area is unusually small.  core.img won't fit in it."));
       goto unable_to_embed;
     }
 
 
-  grub_util_info ("will embed the core image at sector 0x%llx", embed_region.start);
+  grub_util_info ("the core image will be embedded at sector 0x%llx", embed_region.start);
 
   *install_dos_part = grub_cpu_to_le32 (dos_part);
   *install_bsd_part = grub_cpu_to_le32 (bsd_part);
 
   /* The first blocklist contains the whole sectors.  */
   first_block->start = grub_cpu_to_le64 (embed_region.start + 1);
-  first_block->len = grub_cpu_to_le16 (core_sectors - 1);
-  first_block->segment
-    = grub_cpu_to_le16 (GRUB_BOOT_MACHINE_KERNEL_SEG
-			+ (GRUB_DISK_SECTOR_SIZE >> 4));
+
+  /* These are filled elsewhere.  Verify them just in case.  */
+  assert (first_block->len == grub_host_to_target16 (core_sectors - 1));
+  assert (first_block->segment == grub_host_to_target16 (GRUB_BOOT_MACHINE_KERNEL_SEG
+						    + (GRUB_DISK_SECTOR_SIZE >> 4)));
 
   /* Make sure that the second blocklist is a terminator.  */
   block = first_block - 1;
@@ -399,18 +413,20 @@ setup (const char *dir,
 unable_to_embed:
 
   if (must_embed)
-    grub_util_error ("Embedding is not possible, but this is required when "
-		     "the root device is on a RAID array or LVM volume.");
+    grub_util_error (_("embedding is not possible, but this is required when "
+		       "the root device is on a RAID array or LVM volume"));
 
-  grub_util_warn ("Embedding is not possible.  GRUB can only be installed in this "
-		  "setup by using blocklists.  However, blocklists are UNRELIABLE and "
-		  "its use is discouraged.");
+  grub_util_warn (_("Embedding is not possible.  GRUB can only be installed in this "
+		    "setup by using blocklists.  However, blocklists are UNRELIABLE and "
+		    "its use is discouraged."));
   if (! force)
-    grub_util_error ("If you really want blocklists, use --force.");
+    grub_util_error (_("if you really want blocklists, use --force"));
 
   /* Make sure that GRUB reads the identical image as the OS.  */
   tmp_img = xmalloc (core_size);
-  core_path_dev = grub_util_get_path (dir, core_file);
+  core_path_dev_full = grub_util_get_path (dir, core_file);
+  core_path_dev = make_system_path_relative_to_its_root (core_path_dev_full);
+  free (core_path_dev_full);
 
   /* It is a Good Thing to sync two times.  */
   sync ();
@@ -420,8 +436,9 @@ unable_to_embed:
 
   for (i = 0; i < MAX_TRIES; i++)
     {
-      grub_util_info ("attempting to read the core image `%s' from GRUB%s",
-		      core_path_dev, (i == 0) ? "" : " again");
+      grub_util_info ((i == 0) ? _("attempting to read the core image `%s' from GRUB")
+		      : _("attempting to read the core image `%s' from GRUB again"),
+		      core_path_dev);
 
       grub_disk_cache_invalidate_all ();
 
@@ -478,7 +495,7 @@ unable_to_embed:
     }
 
   if (i == MAX_TRIES)
-    grub_util_error ("Cannot read `%s' correctly", core_path_dev);
+    grub_util_error (_("cannot read `%s' correctly"), core_path_dev);
 
   /* Clean out the blocklists.  */
   block = first_block;
@@ -491,7 +508,7 @@ unable_to_embed:
       block--;
 
       if ((char *) block <= core_img)
-	grub_util_error ("No terminator in the core image");
+	grub_util_error (_("no terminator in the core image"));
     }
 
   /* Now read the core image to determine where the sectors are.  */
@@ -502,13 +519,13 @@ unable_to_embed:
   file->read_hook = save_first_sector;
   if (grub_file_read (file, tmp_img, GRUB_DISK_SECTOR_SIZE)
       != GRUB_DISK_SECTOR_SIZE)
-    grub_util_error ("Failed to read the first sector of the core image");
+    grub_util_error (_("failed to read the first sector of the core image"));
 
   block = first_block;
   file->read_hook = save_blocklists;
   if (grub_file_read (file, tmp_img, core_size - GRUB_DISK_SECTOR_SIZE)
       != (grub_ssize_t) core_size - GRUB_DISK_SECTOR_SIZE)
-    grub_util_error ("Failed to read the rest sectors of the core image");
+    grub_util_error (_("failed to read the rest sectors of the core image"));
 
   grub_file_close (file);
 
@@ -527,7 +544,7 @@ unable_to_embed:
   grub_util_info ("opening the core image `%s'", core_path);
   fp = fopen (core_path, "r+b");
   if (! fp)
-    grub_util_error ("Cannot open `%s'", core_path);
+    grub_util_error (_("cannot open `%s'"), core_path);
 
   grub_util_write_image (core_img, GRUB_DISK_SECTOR_SIZE * 2, fp);
   fclose (fp);
@@ -556,6 +573,7 @@ static struct option options[] =
     {"device-map", required_argument, 0, 'm'},
     {"root-device", required_argument, 0, 'r'},
     {"force", no_argument, 0, 'f'},
+    {"skip-fs-probe", no_argument, 0, 's'},
     {"help", no_argument, 0, 'h'},
     {"version", no_argument, 0, 'V'},
     {"verbose", no_argument, 0, 'v'},
@@ -566,13 +584,13 @@ static void
 usage (int status)
 {
   if (status)
-    fprintf (stderr, "Try ``grub-setup --help'' for more information.\n");
+    fprintf (stderr, _("Try `%s --help' for more information.\n"), program_name);
   else
-    printf ("\
-Usage: grub-setup [OPTION]... DEVICE\n\
+    printf (_("\
+Usage: %s [OPTION]... DEVICE\n\
 \n\
 Set up images to boot from DEVICE.\n\
-DEVICE must be a GRUB device (e.g. ``(hd0,1)'').\n\
+DEVICE must be a GRUB device (e.g. `(hd0,1)').\n\
 \n\
   -b, --boot-image=FILE   use FILE as the boot image [default=%s]\n\
   -c, --core-image=FILE   use FILE as the core image [default=%s]\n\
@@ -580,12 +598,14 @@ DEVICE must be a GRUB device (e.g. ``(hd0,1)'').\n\
   -m, --device-map=FILE   use FILE as the device map [default=%s]\n\
   -r, --root-device=DEV   use DEV as the root device [default=guessed]\n\
   -f, --force             install even if problems are detected\n\
+  -s, --skip-fs-probe     do not probe for filesystems in DEVICE\n\
   -h, --help              display this message and exit\n\
   -V, --version           print version information and exit\n\
   -v, --verbose           print verbose messages\n\
 \n\
 Report bugs to <%s>.\n\
-",
+"),
+	    program_name,
 	    DEFAULT_BOOT_FILE, DEFAULT_CORE_FILE, DEFAULT_DIRECTORY,
 	    DEFAULT_DEVICE_MAP, PACKAGE_BUGREPORT);
 
@@ -613,9 +633,11 @@ main (int argc, char *argv[])
   char *dev_map = 0;
   char *root_dev = 0;
   char *dest_dev;
-  int must_embed = 0, force = 0;
+  int must_embed = 0, force = 0, fs_probe = 1;
 
-  progname = "grub-setup";
+  set_program_name (argv[0]);
+
+  grub_util_init_nls ();
 
   /* Check for options.  */
   while (1)
@@ -666,6 +688,10 @@ main (int argc, char *argv[])
 	    force = 1;
 	    break;
 
+	  case 's':
+	    fs_probe = 0;
+	    break;
+
 	  case 'h':
 	    usage (0);
 	    break;
@@ -690,13 +716,13 @@ main (int argc, char *argv[])
   /* Obtain DEST_DEV.  */
   if (optind >= argc)
     {
-      fprintf (stderr, "No device is specified.\n");
+      fprintf (stderr, _("No device is specified.\n"));
       usage (1);
     }
 
   if (optind + 1 != argc)
     {
-      fprintf (stderr, "Unknown extra argument `%s'.\n", argv[optind + 1]);
+      fprintf (stderr, _("Unknown extra argument `%s'.\n"), argv[optind + 1]);
       usage (1);
     }
 
@@ -713,7 +739,7 @@ main (int argc, char *argv[])
       dest_dev = grub_util_get_grub_dev (argv[optind]);
       if (! dest_dev)
 	{
-	  fprintf (stderr, "Invalid device `%s'.\n", argv[optind]);
+	  fprintf (stderr, _("Invalid device `%s'.\n"), argv[optind]);
 	  usage (1);
 	}
     }
@@ -726,7 +752,7 @@ main (int argc, char *argv[])
       char *tmp = get_device_name (root_dev);
 
       if (! tmp)
-	grub_util_error ("Invalid root device `%s'", root_dev);
+	grub_util_error (_("invalid root device `%s'"), root_dev);
 
       tmp = xstrdup (tmp);
       free (root_dev);
@@ -739,7 +765,7 @@ main (int argc, char *argv[])
 	{
 	  grub_util_info ("guessing the root device failed, because of `%s'",
 			  grub_errmsg);
-	  grub_util_error ("Cannot guess the root device. Specify the option ``--root-device''.");
+	  grub_util_error (_("cannot guess the root device. Specify the option `--root-device'"));
 	}
     }
 
@@ -767,7 +793,7 @@ main (int argc, char *argv[])
 	  setup (dir ? : DEFAULT_DIRECTORY,
 		 boot_file ? : DEFAULT_BOOT_FILE,
 		 core_file ? : DEFAULT_CORE_FILE,
-		 root_dev, grub_util_get_grub_dev (devicelist[i]), 1, force);
+		 root_dev, grub_util_get_grub_dev (devicelist[i]), 1, force, fs_probe);
 	}
     }
   else
@@ -776,7 +802,7 @@ main (int argc, char *argv[])
     setup (dir ? : DEFAULT_DIRECTORY,
 	   boot_file ? : DEFAULT_BOOT_FILE,
 	   core_file ? : DEFAULT_CORE_FILE,
-	   root_dev, dest_dev, must_embed, force);
+	   root_dev, dest_dev, must_embed, force, fs_probe);
 
   /* Free resources.  */
   grub_fini_all ();
