@@ -1,7 +1,7 @@
 /* raid.c - RAID support for GRUB utils.  */
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 2006,2007  Free Software Foundation, Inc.
+ *  Copyright (C) 2006,2007,2008  Free Software Foundation, Inc.
  *
  *  GRUB is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -50,8 +50,8 @@ grub_util_getdiskname (int major, int minor)
   else if (major == SCSI_DISK0_MAJOR)
     sprintf (name, "/dev/sd%c", 'a' + minor / 16);
   else
-    grub_util_error ("Unknown device number: %d, %d", major, minor);
-    
+    grub_util_error ("unknown device number: %d, %d", major, minor);
+
   return name;
 }
 
@@ -72,7 +72,7 @@ grub_util_raid_getmembers (char *name)
   fd = open (devname, O_RDONLY);
 
   if (fd == -1)
-    grub_util_error ("Can't open %s: %s", devname, strerror (errno));
+    grub_util_error ("can't open %s: %s", devname, strerror (errno));
 
   free (devname);
 
@@ -81,9 +81,9 @@ grub_util_raid_getmembers (char *name)
     grub_util_error ("ioctl RAID_VERSION error: %s", strerror (errno));
 
   if (version.major != 0 || version.minor != 90)
-    grub_util_error ("Unsupported RAID version: %d.%d",
+    grub_util_error ("unsupported RAID version: %d.%d",
 		     version.major, version.minor);
-  
+
   ret = ioctl (fd, GET_ARRAY_INFO, &info);
   if (ret != 0)
     grub_util_error ("ioctl GET_ARRAY_INFO error: %s", strerror (errno));
@@ -105,7 +105,7 @@ grub_util_raid_getmembers (char *name)
     }
 
   devicelist[j] = NULL;
-  
+
   return devicelist;
 }
 
