@@ -22,14 +22,14 @@
 #include <grub/dl.h>
 #include <grub/mm.h>
 #include <grub/misc.h>
-#include <grub/arg.h>
-#include <grub/file.h>
+#include <grub/bufio.h>
 
 /* Uncomment following define to enable TGA debug.  */
 //#define TGA_DEBUG
 
 #if defined(TGA_DEBUG)
 #define dump_int_field(x) grub_printf( #x " = %d (0x%04x)\n", x, x);
+static grub_command_t cmd;
 #endif
 
 enum
@@ -99,7 +99,7 @@ tga_load_truecolor_rle_R8G8B8 (struct grub_video_bitmap *bitmap,
 
       for (x = 0; x < header->image_width;)
         {
-          if (grub_file_read (file, (char *)&type, sizeof (type)) != sizeof(type))
+          if (grub_file_read (file, &type, sizeof (type)) != sizeof(type))
             return grub_errno;
 
           if (type & 0x80)
@@ -108,7 +108,7 @@ tga_load_truecolor_rle_R8G8B8 (struct grub_video_bitmap *bitmap,
               type &= 0x7f;
               type++;
 
-              if (grub_file_read (file, (char *)&tmp[0], bytes_per_pixel)
+              if (grub_file_read (file, &tmp[0], bytes_per_pixel)
                   != bytes_per_pixel)
                 return grub_errno;
 
@@ -133,7 +133,7 @@ tga_load_truecolor_rle_R8G8B8 (struct grub_video_bitmap *bitmap,
 
               while (type)
                 {
-                  if (grub_file_read (file, (char *)&tmp[0], bytes_per_pixel)
+                  if (grub_file_read (file, &tmp[0], bytes_per_pixel)
                       != bytes_per_pixel)
                     return grub_errno;
 
@@ -178,7 +178,7 @@ tga_load_truecolor_rle_R8G8B8A8 (struct grub_video_bitmap *bitmap,
 
       for (x = 0; x < header->image_width;)
         {
-          if (grub_file_read (file, (char *)&type, sizeof (type)) != sizeof(type))
+          if (grub_file_read (file, &type, sizeof (type)) != sizeof(type))
             return grub_errno;
 
           if (type & 0x80)
@@ -187,7 +187,7 @@ tga_load_truecolor_rle_R8G8B8A8 (struct grub_video_bitmap *bitmap,
               type &= 0x7f;
               type++;
 
-              if (grub_file_read (file, (char *)&tmp[0], bytes_per_pixel)
+              if (grub_file_read (file, &tmp[0], bytes_per_pixel)
                   != bytes_per_pixel)
                 return grub_errno;
 
@@ -213,7 +213,7 @@ tga_load_truecolor_rle_R8G8B8A8 (struct grub_video_bitmap *bitmap,
 
               while (type)
                 {
-                  if (grub_file_read (file, (char *)&tmp[0], bytes_per_pixel)
+                  if (grub_file_read (file, &tmp[0], bytes_per_pixel)
                       != bytes_per_pixel)
                     return grub_errno;
 
@@ -258,7 +258,7 @@ tga_load_truecolor_R8G8B8 (struct grub_video_bitmap *bitmap,
 
       for (x = 0; x < header->image_width; x++)
         {
-          if (grub_file_read (file, (char *)&tmp[0], bytes_per_pixel)
+          if (grub_file_read (file, &tmp[0], bytes_per_pixel)
               != bytes_per_pixel)
             return grub_errno;
 
@@ -295,7 +295,7 @@ tga_load_truecolor_R8G8B8A8 (struct grub_video_bitmap *bitmap,
 
       for (x = 0; x < header->image_width; x++)
         {
-          if (grub_file_read (file, (char *)&tmp[0], bytes_per_pixel)
+          if (grub_file_read (file, &tmp[0], bytes_per_pixel)
               != bytes_per_pixel)
             return grub_errno;
 
@@ -319,7 +319,7 @@ grub_video_reader_tga (struct grub_video_bitmap **bitmap,
   struct grub_tga_header header;
   int has_alpha;
 
-  file = grub_file_open (filename);
+  file = grub_buffile_open (filename, 0);
   if (! file)
     return grub_errno;
 
@@ -328,7 +328,7 @@ grub_video_reader_tga (struct grub_video_bitmap **bitmap,
      not going to support developer area & extensions at this point.  */
 
   /* Read TGA header from beginning of file.  */
-  if (grub_file_read (file, (char*)&header, sizeof (header)) 
+  if (grub_file_read (file, &header, sizeof (header))
       != sizeof (header))
     {
       grub_file_close (file);
@@ -371,7 +371,7 @@ grub_video_reader_tga (struct grub_video_bitmap **bitmap,
       default:
         grub_file_close (file);
         return grub_error (GRUB_ERR_BAD_FILE_TYPE,
-                           "Unsupported bitmap format (unknown encoding).");
+                           "unsupported bitmap format (unknown encoding)");
     }
 
   /* Check that bitmap depth is supported.  */
@@ -388,7 +388,7 @@ grub_video_reader_tga (struct grub_video_bitmap **bitmap,
       default:
         grub_file_close (file);
         return grub_error (GRUB_ERR_BAD_FILE_TYPE,
-                           "Unsupported bitmap format (bpp=%d).",
+                           "unsupported bitmap format (bpp=%d)",
                            header.image_bpp);
     }
 
@@ -397,7 +397,7 @@ grub_video_reader_tga (struct grub_video_bitmap **bitmap,
     {
       grub_video_bitmap_create (bitmap, header.image_width,
                                 header.image_height,
-                                GRUB_VIDEO_BLIT_FORMAT_R8G8B8A8);
+                                GRUB_VIDEO_BLIT_FORMAT_RGBA_8888);
       if (grub_errno != GRUB_ERR_NONE)
         {
           grub_file_close (file);
@@ -420,7 +420,7 @@ grub_video_reader_tga (struct grub_video_bitmap **bitmap,
     {
       grub_video_bitmap_create (bitmap, header.image_width,
                                 header.image_height,
-                                GRUB_VIDEO_BLIT_FORMAT_R8G8B8);
+                                GRUB_VIDEO_BLIT_FORMAT_RGB_888);
       if (grub_errno != GRUB_ERR_NONE)
         {
           grub_file_close (file);
@@ -453,7 +453,7 @@ grub_video_reader_tga (struct grub_video_bitmap **bitmap,
 
 #if defined(TGA_DEBUG)
 static grub_err_t
-grub_cmd_tgatest (struct grub_arg_list *state __attribute__ ((unused)),
+grub_cmd_tgatest (grub_command_t cmd __attribute__ ((unused)),
                   int argc, char **args)
 {
   struct grub_video_bitmap *bitmap = 0;
@@ -461,7 +461,7 @@ grub_cmd_tgatest (struct grub_arg_list *state __attribute__ ((unused)),
   if (argc != 1)
     return grub_error (GRUB_ERR_BAD_ARGUMENT, "file name required");
 
-  grub_video_reader_tga (&bitmap, args[0]);  
+  grub_video_reader_tga (&bitmap, args[0]);
   if (grub_errno != GRUB_ERR_NONE)
     return grub_errno;
 
@@ -481,15 +481,15 @@ GRUB_MOD_INIT(video_reader_tga)
 {
   grub_video_bitmap_reader_register (&tga_reader);
 #if defined(TGA_DEBUG)
-  grub_register_command ("tgatest", grub_cmd_tgatest, GRUB_COMMAND_FLAG_BOTH,
-                         "tgatest FILE", "Tests loading of TGA bitmap.", 0);
+  cmd = grub_register_command ("tgatest", grub_cmd_tgatest,
+                               "FILE", "Tests loading of TGA bitmap.");
 #endif
 }
 
 GRUB_MOD_FINI(video_reader_tga)
 {
 #if defined(TGA_DEBUG)
-  grub_unregister_command ("tgatest");
+  grub_unregister_command (cmd);
 #endif
   grub_video_bitmap_reader_unregister (&tga_reader);
 }
