@@ -81,7 +81,7 @@ grub_err_t
 grub_video_get_info (struct grub_video_mode_info *mode_info)
 {
   if (! grub_video_adapter_active)
-    return grub_error (GRUB_ERR_BAD_DEVICE, "No video mode activated");
+    return grub_error (GRUB_ERR_BAD_DEVICE, "no video mode activated");
 
   /* If mode_info is NULL just report that video adapter is active.  */
   if (! mode_info)
@@ -93,6 +93,32 @@ grub_video_get_info (struct grub_video_mode_info *mode_info)
   return grub_video_adapter_active->get_info (mode_info);
 }
 
+grub_video_driver_id_t
+grub_video_get_driver_id (void)
+{
+  if (! grub_video_adapter_active)
+    return GRUB_VIDEO_DRIVER_NONE;
+  return grub_video_adapter_active->id;
+}
+
+/* Get information about active video mode.  */
+grub_err_t
+grub_video_get_info_and_fini (struct grub_video_mode_info *mode_info,
+			      void **framebuffer)
+{
+  grub_err_t err;
+
+  if (! grub_video_adapter_active)
+    return grub_error (GRUB_ERR_BAD_DEVICE, "no video mode activated");
+
+  err = grub_video_adapter_active->get_info_and_fini (mode_info, framebuffer);
+  if (err)
+    return err;
+
+  grub_video_adapter_active = 0;
+  return GRUB_ERR_NONE;
+}
+
 /* Determine optimized blitting formation for specified video mode info.  */
 enum grub_video_blit_format
 grub_video_get_blit_format (struct grub_video_mode_info *mode_info)
@@ -163,6 +189,8 @@ grub_video_get_blit_format (struct grub_video_mode_info *mode_info)
 	  return GRUB_VIDEO_BLIT_FORMAT_RGB_565;
 	}
     }
+  else if (mode_info->bpp == 1)
+    return GRUB_VIDEO_BLIT_FORMAT_1BIT_PACKED;
 
   /* Backup route.  Unknown format.  */
 
@@ -189,7 +217,7 @@ grub_video_set_palette (unsigned int start, unsigned int count,
                         struct grub_video_palette_data *palette_data)
 {
   if (! grub_video_adapter_active)
-    return grub_error (GRUB_ERR_BAD_DEVICE, "No video mode activated");
+    return grub_error (GRUB_ERR_BAD_DEVICE, "no video mode activated");
 
   return grub_video_adapter_active->set_palette (start, count, palette_data);
 }
@@ -200,7 +228,7 @@ grub_video_get_palette (unsigned int start, unsigned int count,
                         struct grub_video_palette_data *palette_data)
 {
   if (! grub_video_adapter_active)
-    return grub_error (GRUB_ERR_BAD_DEVICE, "No video mode activated");
+    return grub_error (GRUB_ERR_BAD_DEVICE, "no video mode activated");
 
   return grub_video_adapter_active->get_palette (start, count, palette_data);
 }
@@ -211,7 +239,7 @@ grub_video_set_viewport (unsigned int x, unsigned int y,
                          unsigned int width, unsigned int height)
 {
   if (! grub_video_adapter_active)
-    return grub_error (GRUB_ERR_BAD_DEVICE, "No video mode activated");
+    return grub_error (GRUB_ERR_BAD_DEVICE, "no video mode activated");
 
   return grub_video_adapter_active->set_viewport (x, y, width, height);
 }
@@ -222,7 +250,7 @@ grub_video_get_viewport (unsigned int *x, unsigned int *y,
                          unsigned int *width, unsigned int *height)
 {
   if (! grub_video_adapter_active)
-    return grub_error (GRUB_ERR_BAD_DEVICE, "No video mode activated");
+    return grub_error (GRUB_ERR_BAD_DEVICE, "no video mode activated");
 
   return grub_video_adapter_active->get_viewport (x, y, width, height);
 }
@@ -265,7 +293,7 @@ grub_video_unmap_color (grub_video_color_t color, grub_uint8_t *red,
                         grub_uint8_t *alpha)
 {
   if (! grub_video_adapter_active)
-    return grub_error (GRUB_ERR_BAD_DEVICE, "No video mode activated");
+    return grub_error (GRUB_ERR_BAD_DEVICE, "no video mode activated");
 
   return grub_video_adapter_active->unmap_color (color,
                                                  red,
@@ -280,7 +308,7 @@ grub_video_fill_rect (grub_video_color_t color, int x, int y,
                       unsigned int width, unsigned int height)
 {
   if (! grub_video_adapter_active)
-    return grub_error (GRUB_ERR_BAD_DEVICE, "No video mode activated");
+    return grub_error (GRUB_ERR_BAD_DEVICE, "no video mode activated");
 
   return grub_video_adapter_active->fill_rect (color, x, y, width, height);
 }
@@ -293,7 +321,7 @@ grub_video_blit_bitmap (struct grub_video_bitmap *bitmap,
                         unsigned int width, unsigned int height)
 {
   if (! grub_video_adapter_active)
-    return grub_error (GRUB_ERR_BAD_DEVICE, "No video mode activated");
+    return grub_error (GRUB_ERR_BAD_DEVICE, "no video mode activated");
 
   return grub_video_adapter_active->blit_bitmap (bitmap, oper, x, y,
                                                  offset_x, offset_y,
@@ -308,7 +336,7 @@ grub_video_blit_render_target (struct grub_video_render_target *target,
                                unsigned int width, unsigned int height)
 {
   if (! grub_video_adapter_active)
-    return grub_error (GRUB_ERR_BAD_DEVICE, "No video mode activated");
+    return grub_error (GRUB_ERR_BAD_DEVICE, "no video mode activated");
 
   return grub_video_adapter_active->blit_render_target (target, oper, x, y,
                                                         offset_x, offset_y,
@@ -320,7 +348,7 @@ grub_err_t
 grub_video_scroll (grub_video_color_t color, int dx, int dy)
 {
   if (! grub_video_adapter_active)
-    return grub_error (GRUB_ERR_BAD_DEVICE, "No video mode activated");
+    return grub_error (GRUB_ERR_BAD_DEVICE, "no video mode activated");
 
   return grub_video_adapter_active->scroll (color, dx, dy);
 }
@@ -330,7 +358,7 @@ grub_err_t
 grub_video_swap_buffers (void)
 {
   if (! grub_video_adapter_active)
-    return grub_error (GRUB_ERR_BAD_DEVICE, "No video mode activated");
+    return grub_error (GRUB_ERR_BAD_DEVICE, "no video mode activated");
 
   return grub_video_adapter_active->swap_buffers ();
 }
@@ -342,7 +370,7 @@ grub_video_create_render_target (struct grub_video_render_target **result,
                                  unsigned int mode_type)
 {
   if (! grub_video_adapter_active)
-    return grub_error (GRUB_ERR_BAD_DEVICE, "No video mode activated");
+    return grub_error (GRUB_ERR_BAD_DEVICE, "no video mode activated");
 
   return grub_video_adapter_active->create_render_target (result,
                                                           width, height,
@@ -354,7 +382,7 @@ grub_err_t
 grub_video_delete_render_target (struct grub_video_render_target *target)
 {
   if (! grub_video_adapter_active)
-    return grub_error (GRUB_ERR_BAD_DEVICE, "No video mode activated");
+    return grub_error (GRUB_ERR_BAD_DEVICE, "no video mode activated");
 
   return grub_video_adapter_active->delete_render_target (target);
 }
@@ -364,7 +392,7 @@ grub_err_t
 grub_video_set_active_render_target (struct grub_video_render_target *target)
 {
   if (! grub_video_adapter_active)
-    return grub_error (GRUB_ERR_BAD_DEVICE, "No video mode activated");
+    return grub_error (GRUB_ERR_BAD_DEVICE, "no video mode activated");
 
   return grub_video_adapter_active->set_active_render_target (target);
 }
@@ -374,26 +402,86 @@ grub_err_t
 grub_video_get_active_render_target (struct grub_video_render_target **target)
 {
   if (! grub_video_adapter_active)
-    return grub_error (GRUB_ERR_BAD_DEVICE, "No video mode activated");
+    return grub_error (GRUB_ERR_BAD_DEVICE, "no video mode activated");
 
   return grub_video_adapter_active->get_active_render_target (target);
 }
 
+/* Parse <width>x<height>[x<depth>]*/
+static grub_err_t
+parse_modespec (const char *current_mode, int *width, int *height, int *depth)
+{
+  const char *value;
+  const char *param = current_mode;
+
+  *width = *height = *depth = -1;
+
+  if (grub_strcmp (param, "auto") == 0)
+    {
+      *width = *height = 0;
+      return GRUB_ERR_NONE;
+    }
+
+  /* Find width value.  */
+  value = param;
+  param = grub_strchr(param, 'x');
+  if (param == NULL)
+    return grub_error (GRUB_ERR_BAD_ARGUMENT,
+		       "Invalid mode: %s\n",
+		       current_mode);
+
+  param++;
+  
+  *width = grub_strtoul (value, 0, 0);
+  if (grub_errno != GRUB_ERR_NONE)
+      return grub_error (GRUB_ERR_BAD_ARGUMENT,
+			 "Invalid mode: %s\n",
+			 current_mode);
+  
+  /* Find height value.  */
+  value = param;
+  param = grub_strchr(param, 'x');
+  if (param == NULL)
+    {
+      *height = grub_strtoul (value, 0, 0);
+      if (grub_errno != GRUB_ERR_NONE)
+	return grub_error (GRUB_ERR_BAD_ARGUMENT,
+			   "Invalid mode: %s\n",
+			   current_mode);
+    }
+  else
+    {
+      /* We have optional color depth value.  */
+      param++;
+      
+      *height = grub_strtoul (value, 0, 0);
+      if (grub_errno != GRUB_ERR_NONE)
+	return grub_error (GRUB_ERR_BAD_ARGUMENT,
+			   "Invalid mode: %s\n",
+			   current_mode);
+      
+      /* Convert color depth value.  */
+      value = param;
+      *depth = grub_strtoul (value, 0, 0);
+      if (grub_errno != GRUB_ERR_NONE)
+	return grub_error (GRUB_ERR_BAD_ARGUMENT,
+			   "Invalid mode: %s\n",
+			   current_mode);
+    }
+  return GRUB_ERR_NONE;
+}
+
 grub_err_t
-grub_video_set_mode (char *modestring,
-		     int NESTED_FUNC_ATTR (*hook) (grub_video_adapter_t p,
-						   struct grub_video_mode_info *mode_info))
+grub_video_set_mode (const char *modestring,
+		     unsigned int modemask,
+		     unsigned int modevalue)
 {
   char *tmp;
   char *next_mode;
   char *current_mode;
-  char *param;
-  char *value;
   char *modevar;
-  int width = -1;
-  int height = -1;
-  int depth = -1;
-  int flags = 0;
+
+  modevalue &= modemask;
 
   /* Take copy of env.var. as we don't want to modify that.  */
   modevar = grub_strdup (modestring);
@@ -409,26 +497,26 @@ grub_video_set_mode (char *modestring,
       || grub_memcmp (next_mode, "keep,", sizeof ("keep,") - 1) == 0
       || grub_memcmp (next_mode, "keep;", sizeof ("keep;") - 1) == 0)
     {
-      struct grub_video_mode_info mode_info;
       int suitable = 1;
       grub_err_t err;
 
-      grub_memset (&mode_info, 0, sizeof (mode_info));
-
       if (grub_video_adapter_active)
 	{
+	  struct grub_video_mode_info mode_info;
+	  grub_memset (&mode_info, 0, sizeof (mode_info));
 	  err = grub_video_get_info (&mode_info);
 	  if (err)
 	    {
 	      suitable = 0;
 	      grub_errno = GRUB_ERR_NONE;
 	    }
+	  if ((mode_info.mode_type & modemask) != modevalue)
+	    suitable = 0;
 	}
-      else
-	mode_info.mode_type = GRUB_VIDEO_MODE_TYPE_PURE_TEXT;
+      else if (((GRUB_VIDEO_MODE_TYPE_PURE_TEXT & modemask) != 0)
+	       && ((GRUB_VIDEO_MODE_TYPE_PURE_TEXT & modevalue) == 0))
+	suitable = 0;
 
-      if (suitable && hook)
-	suitable = hook (grub_video_adapter_active, &mode_info);
       if (suitable)
 	{
 	  grub_free (modevar);
@@ -440,7 +528,7 @@ grub_video_set_mode (char *modestring,
 	  grub_free (modevar);
 
 	  return grub_error (GRUB_ERR_BAD_ARGUMENT,
-			     "No suitable mode found.");
+			     "no suitable mode found");
 	}
 
       /* Skip separator. */
@@ -462,15 +550,16 @@ grub_video_set_mode (char *modestring,
   /* Loop until all modes has been tested out.  */
   while (next_mode != NULL)
     {
+      int width = -1;
+      int height = -1;
+      int depth = -1;
+      grub_err_t err;
+      unsigned int flags = modevalue;
+      unsigned int flagmask = modemask;
+
       /* Use last next_mode as current mode.  */
       tmp = next_mode;
 
-      /* Reset video mode settings.  */
-      width = -1;
-      height = -1;
-      depth = -1;
-      flags = 0;
-
       /* Save position of next mode and separate modes.  */
       for (; *next_mode; next_mode++)
 	if (*next_mode == ',' || *next_mode == ';')
@@ -489,19 +578,16 @@ grub_video_set_mode (char *modestring,
 
       /* Initialize token holders.  */
       current_mode = tmp;
-      param = tmp;
-      value = NULL;
 
       /* XXX: we assume that we're in pure text mode if
 	 no video mode is initialized. Is it always true? */
-      if (grub_strcmp (param, "text") == 0)
+      if (grub_strcmp (current_mode, "text") == 0)
 	{
 	  struct grub_video_mode_info mode_info;
 
 	  grub_memset (&mode_info, 0, sizeof (mode_info));
-	  mode_info.mode_type = GRUB_VIDEO_MODE_TYPE_PURE_TEXT;
-
-	  if (! hook || hook (0, &mode_info))
+	  if (((GRUB_VIDEO_MODE_TYPE_PURE_TEXT & modemask) == 0)
+	      || ((GRUB_VIDEO_MODE_TYPE_PURE_TEXT & modevalue) != 0))
 	    {
 	      /* Valid mode found from adapter, and it has been activated.
 		 Specify it as active adapter.  */
@@ -514,121 +600,31 @@ grub_video_set_mode (char *modestring,
 	    }
 	}
 
-      /* Parse <width>x<height>[x<depth>]*/
-
-      /* Find width value.  */
-      value = param;
-      param = grub_strchr(param, 'x');
-      if (param == NULL)
+      err = parse_modespec (current_mode, &width, &height, &depth);
+      if (err)
 	{
-	  grub_err_t rc;
-
-	  /* First setup error message.  */
-	  rc = grub_error (GRUB_ERR_BAD_ARGUMENT,
-			   "Invalid mode: %s\n",
-			   current_mode);
-
 	  /* Free memory before returning.  */
 	  grub_free (modevar);
 
-	  return rc;
-	}
-
-      *param = 0;
-      param++;
-
-      width = grub_strtoul (value, 0, 0);
-      if (grub_errno != GRUB_ERR_NONE)
-	{
-	  grub_err_t rc;
-
-	  /* First setup error message.  */
-	  rc = grub_error (GRUB_ERR_BAD_ARGUMENT,
-			   "Invalid mode: %s\n",
-			   current_mode);
-
-	  /* Free memory before returning.  */
-	  grub_free (modevar);
-
-	  return rc;
-	}
-
-      /* Find height value.  */
-      value = param;
-      param = grub_strchr(param, 'x');
-      if (param == NULL)
-	{
-	  height = grub_strtoul (value, 0, 0);
-	  if (grub_errno != GRUB_ERR_NONE)
-	    {
-	      grub_err_t rc;
-
-	      /* First setup error message.  */
-	      rc = grub_error (GRUB_ERR_BAD_ARGUMENT,
-			       "Invalid mode: %s\n",
-			       current_mode);
-
-	      /* Free memory before returning.  */
-	      grub_free (modevar);
-
-	      return rc;
-	    }
-	}
-      else
-	{
-	  /* We have optional color depth value.  */
-	  *param = 0;
-	  param++;
-
-	  height = grub_strtoul (value, 0, 0);
-	  if (grub_errno != GRUB_ERR_NONE)
-	    {
-	      grub_err_t rc;
-
-	      /* First setup error message.  */
-	      rc = grub_error (GRUB_ERR_BAD_ARGUMENT,
-			       "Invalid mode: %s\n",
-			       current_mode);
-
-	      /* Free memory before returning.  */
-	      grub_free (modevar);
-
-	      return rc;
-	    }
-
-	  /* Convert color depth value.  */
-	  value = param;
-	  depth = grub_strtoul (value, 0, 0);
-	  if (grub_errno != GRUB_ERR_NONE)
-	    {
-	      grub_err_t rc;
-
-	      /* First setup error message.  */
-	      rc = grub_error (GRUB_ERR_BAD_ARGUMENT,
-			       "Invalid mode: %s\n",
-			       current_mode);
-
-	      /* Free memory before returning.  */
-	      grub_free (modevar);
-
-	      return rc;
-	    }
+	  return err;
 	}
 
       /* Try out video mode.  */
 
-      /* If we have 8 or less bits, then assume that it is indexed color mode.  */
-      if ((depth <= 8) && (depth != -1))
-	flags |= GRUB_VIDEO_MODE_TYPE_INDEX_COLOR;
+      /* If user requested specific depth check if this depth is supported.  */
+      if (depth != -1 && (flagmask & GRUB_VIDEO_MODE_TYPE_DEPTH_MASK)
+	  &&
+	  (((flags & GRUB_VIDEO_MODE_TYPE_DEPTH_MASK)
+	    != ((depth << GRUB_VIDEO_MODE_TYPE_DEPTH_POS)
+		& GRUB_VIDEO_MODE_TYPE_DEPTH_MASK))))
+	continue;
 
-      /* We have more than 8 bits, then assume that it is RGB color mode.  */
-      if (depth > 8)
-	flags |= GRUB_VIDEO_MODE_TYPE_RGB;
-
-      /* If user requested specific depth, forward that information to driver.  */
       if (depth != -1)
-	flags |= (depth << GRUB_VIDEO_MODE_TYPE_DEPTH_POS)
-	  & GRUB_VIDEO_MODE_TYPE_DEPTH_MASK;
+	{
+	  flags |= (depth << GRUB_VIDEO_MODE_TYPE_DEPTH_POS)
+	    & GRUB_VIDEO_MODE_TYPE_DEPTH_MASK;
+	  flagmask |= GRUB_VIDEO_MODE_TYPE_DEPTH_MASK;
+	}
 
       /* Try to initialize requested mode.  Ignore any errors.  */
       grub_video_adapter_t p;
@@ -636,7 +632,6 @@ grub_video_set_mode (char *modestring,
       /* Loop thru all possible video adapter trying to find requested mode.  */
       for (p = grub_video_adapter_list; p; p = p->next)
 	{
-	  grub_err_t err;
 	  struct grub_video_mode_info mode_info;
 
 	  grub_memset (&mode_info, 0, sizeof (mode_info));
@@ -650,7 +645,7 @@ grub_video_set_mode (char *modestring,
 	    }
 
 	  /* Try to initialize video mode.  */
-	  err = p->setup (width, height, flags);
+	  err = p->setup (width, height, flags, flagmask);
 	  if (err != GRUB_ERR_NONE)
 	    {
 	      p->fini ();
@@ -666,7 +661,15 @@ grub_video_set_mode (char *modestring,
 	      continue;
 	    }
 
-	  if (hook && ! hook (p, &mode_info))
+	  flags = mode_info.mode_type & ~GRUB_VIDEO_MODE_TYPE_DEPTH_MASK;
+	  flags |= (mode_info.bpp << GRUB_VIDEO_MODE_TYPE_DEPTH_POS)
+	    & GRUB_VIDEO_MODE_TYPE_DEPTH_MASK;
+
+	  /* Check that mode is suitable for upper layer.  */
+	  if ((flags & GRUB_VIDEO_MODE_TYPE_PURE_TEXT)
+	      ? (((GRUB_VIDEO_MODE_TYPE_PURE_TEXT & modemask) != 0)
+		 && ((GRUB_VIDEO_MODE_TYPE_PURE_TEXT & modevalue) == 0))
+	      : ((flags & modemask) != modevalue))
 	    {
 	      p->fini ();
 	      grub_errno = GRUB_ERR_NONE;
@@ -689,14 +692,12 @@ grub_video_set_mode (char *modestring,
   grub_free (modevar);
 
   return grub_error (GRUB_ERR_BAD_ARGUMENT,
-		     "No suitable mode found.");
+		     "no suitable mode found");
 }
 
 /* Initialize Video API module.  */
 GRUB_MOD_INIT(video_video)
 {
-  grub_video_adapter_active = 0;
-  grub_video_adapter_list = 0;
 }
 
 /* Finalize Video API module.  */
